import math
import warnings
from typing import List, Literal, Tuple, TypeAlias

import torch

from gfn.containers import Trajectories
from gfn.env import Env
from gfn.estimators import ConditionalScalarEstimator, Estimator, ScalarEstimator
from gfn.gflownet.base import TrajectoryBasedGFlowNet, loss_reduce
from gfn.utils.handlers import (
    has_conditioning_exception_handler,
    no_conditioning_exception_handler,
    warn_about_recalculating_logprobs,
)

ContributionsTensor: TypeAlias = (
    torch.Tensor
)  # shape: [maxlen * (1 + maxlen) / 2, n_traj]
CumulativeLogProbsTensor: TypeAlias = torch.Tensor  # shape: [maxlen + 1, n_traj]
LogStateFlowsTensor: TypeAlias = torch.Tensor  # shape: [maxlen, n_traj]
LogTrajectoriesTensor: TypeAlias = torch.Tensor  # shape: [maxlen, n_traj]
MaskTensor: TypeAlias = torch.Tensor  # shape: [maxlen, n_traj]
PredictionsTensor: TypeAlias = torch.Tensor  # shape: [maxlen + 1 - i, n_traj]
TargetsTensor: TypeAlias = torch.Tensor  # shape: [maxlen + 1 - i, n_traj]


class SubTBGFlowNet(TrajectoryBasedGFlowNet):
    r"""GFlowNet for the Sub-Trajectory Balance loss.

    An implementation of the sub-trajectory balance loss as described in
    [Learning GFlowNets from partial episodes for improved convergence and stability](https://arxiv.org/abs/2209.12782).

    Attributes:
        pf: The forward policy estimator.
        pb: The backward policy estimator, or None if the gflownet DAG is a tree, and
            pb is therefore always 1.
        logF: A ScalarEstimator or ConditionalScalarEstimator for estimating the log flow
            of the states.
        weighting: The sub-trajectories weighting scheme.
            - "DB": Considers all one-step transitions of each trajectory in the
                batch and weighs them equally (regardless of the length of
                trajectory). Should be equivalent to DetailedBalance loss.
            - "ModifiedDB": Considers all one-step transitions of each trajectory
                in the batch and weighs them inversely proportional to the
                trajectory length. This ensures that the loss is not dominated by
                long trajectories. Each trajectory contributes equally to the loss.
            - "TB": Considers only the full trajectory. Should be equivalent to
                TrajectoryBalance loss.
            - "equal_within": Each sub-trajectory of each trajectory is weighed
                equally within the trajectory. Then each trajectory is weighed
                equally within the batch.
            - "equal": Each sub-trajectory of each trajectory is weighed equally
                within the set of all sub-trajectories.
            - "geometric_within": Each sub-trajectory of each trajectory is weighed
                proportionally to (lamda ** len(sub_trajectory)), within each
                trajectory. THIS CORRESPONDS TO THE ONE IN THE PAPER.
            - "geometric": Each sub-trajectory of each trajectory is weighed
                proportionally to (lamda ** len(sub_trajectory)), within the set of
                all sub-trajectories.
        lamda: Discount factor for longer trajectories (used in geometric weighting).
        log_reward_clip_min: If finite, clips log rewards to this value.
        forward_looking: Whether to use the forward-looking GFN loss.
        constant_pb: Whether to ignore the backward policy estimator, e.g., if the
            gflownet DAG is a tree, and pb is therefore always 1.
    """

    def __init__(
        self,
        pf: Estimator,
        pb: Estimator | None,
        logF: ScalarEstimator | ConditionalScalarEstimator,
        weighting: Literal[
            "DB",
            "ModifiedDB",
            "TB",
            "geometric",
            "equal",
            "geometric_within",
            "equal_within",
        ] = "geometric_within",
        lamda: float = 0.9,
        log_reward_clip_min: float = -float("inf"),
        forward_looking: bool = False,
        constant_pb: bool = False,
    ):
        """Initializes a SubTBGFlowNet instance.

        Args:
            pf: The forward policy estimator.
            pb: The backward policy estimator.
            logF: A ScalarEstimator or ConditionalScalarEstimator for estimating the
                log flow of the states.
            weighting: The sub-trajectory weighting scheme (see class docstring for
                details).
            lamda: Discount factor for longer trajectories (used in geometric weighting).
            log_reward_clip_min: If finite, clips log rewards to this value.
            forward_looking: Whether to use the forward-looking GFN loss.
            constant_pb: Whether to ignore the backward policy estimator, e.g., if the
                gflownet DAG is a tree, and pb is therefore always 1. Must be set
                explicitly by user to ensure that pb is an Estimator except under this
                special case.
        """
        super().__init__(pf, pb, constant_pb=constant_pb)
        assert any(
            isinstance(logF, cls)
            for cls in [ScalarEstimator, ConditionalScalarEstimator]
        ), "logF must be a ScalarEstimator or derived"
        self.logF = logF
        self.weighting = weighting
        self.lamda = lamda
        self.log_reward_clip_min = log_reward_clip_min
        self.forward_looking = forward_looking

    def logF_named_parameters(self) -> dict[str, torch.Tensor]:
        """Returns a dictionary of named parameters containing 'logF' in their name.

        Returns:
            A dictionary of named parameters containing 'logF' in their name.
        """
        return {k: v for k, v in self.named_parameters() if "logF" in k}

    def logF_parameters(self) -> list[torch.Tensor]:
        """Returns a list of parameters containing 'logF' in their name.

        Returns:
            A list of parameters containing 'logF' in their name.
        """
        return [v for k, v in self.named_parameters() if "logF" in k]

    def cumulative_logprobs(
        self,
        trajectories: Trajectories,
        log_p_trajectories: LogTrajectoriesTensor,
    ) -> CumulativeLogProbsTensor:
        """Calculates the cumulative logprobs for all trajectories.

        Args:
            trajectories: The batch of trajectories.
            log_p_trajectories: Tensor of shape (max_length, n_trajectories)
                containing the logprobs of the forward or backward actions of the
                trajectories.

        Returns:
            A tensor of shape (max_length + 1, n_trajectories) containing the
            cumulative sum of logprobs for each trajectory.
        """
        return torch.cat(
            (
                torch.zeros(
                    1, trajectories.n_trajectories, device=log_p_trajectories.device
                ),
                log_p_trajectories.cumsum(dim=0),
            ),
            dim=0,
        )

    def calculate_preds(
        self,
        log_pf_trajectories_cum: CumulativeLogProbsTensor,
        log_state_flows: LogStateFlowsTensor,
        i: int,
    ) -> PredictionsTensor:
        """Calculates the predictions tensor for the current sub-trajectory length.

        Args:
            log_pf_trajectories_cum: Tensor of shape (max_length + 1, n_trajectories)
                containing the cumulative sum of logprobs of the forward actions for
                each trajectory.
            log_state_flows: Tensor of shape (max_length, n_trajectories) containing
                the estimated log flow of the states.
            i: The sub-trajectory length.

        Returns:
            The predictions tensor of shape (max_length + 1 - i, n_trajectories).
        """
        current_log_state_flows = (
            log_state_flows if i == 1 else log_state_flows[: -(i - 1)]
        )

        preds = (
            log_pf_trajectories_cum[i:]
            - log_pf_trajectories_cum[:-i]
            + current_log_state_flows
        )

        return preds

    def calculate_targets(
        self,
        trajectories: Trajectories,
        preds: PredictionsTensor,
        log_pb_trajectories_cum: CumulativeLogProbsTensor,
        log_state_flows: LogStateFlowsTensor,
        is_terminal_mask: MaskTensor,
        sink_states_mask: MaskTensor,
        i: int,
    ) -> TargetsTensor:
        """Calculates the targets tensor for the current sub-trajectory length.

        Args:
            trajectories: The batch of trajectories.
            preds: Tensor of shape (max_length + 1 - i, n_trajectories) containing
                the predictions for the current sub-trajectory length.
            log_pb_trajectories_cum: Tensor of shape (max_length + 1, n_trajectories)
                containing the cumulative sum of logprobs of the backward actions for
                each trajectory.
            log_state_flows: Tensor of shape (max_length, n_trajectories) containing
                the estimated log flow of the states.
            is_terminal_mask: A mask of shape (max_length, n_trajectories) indicating
                whether the state is terminal.
            sink_states_mask: A mask of shape (max_length, n_trajectories) indicating
                whether the state is a sink state.
            i: The sub-trajectory length.

        Returns:
            The targets tensor of shape (max_length + 1 - i, n_trajectories).
        """
        targets = torch.full_like(preds, fill_value=-float("inf"))
        assert trajectories.log_rewards is not None
        log_rewards = trajectories.log_rewards[trajectories.terminating_idx >= i]

        if math.isfinite(self.log_reward_clip_min):
            log_rewards.clamp_min(self.log_reward_clip_min)

        targets.T[is_terminal_mask[i - 1 :].T] = log_rewards

        # For now, the targets contain the log-rewards of the ending sub trajectories
        # We need to add to that the log-probabilities of the backward actions up-to
        # the sub-trajectory's terminating state
        if i > 1:
            targets[is_terminal_mask[i - 1 :]] += (
                log_pb_trajectories_cum[i - 1 :] - log_pb_trajectories_cum[: -i + 1]
            )[:-1][is_terminal_mask[i - 1 :]]

        # The following creates the targets for the non-finishing sub-trajectories
        full_mask = sink_states_mask | is_terminal_mask
        targets[~full_mask[i - 1 :]] = (
            log_pb_trajectories_cum[i:] - log_pb_trajectories_cum[:-i]
        )[:-1][~full_mask[i - 1 : -1]] + log_state_flows[i:][~sink_states_mask[i:]]

        return targets

    def calculate_log_state_flows(
        self,
        env: Env,
        trajectories: Trajectories,
        log_pf_trajectories: LogTrajectoriesTensor,
    ) -> LogStateFlowsTensor:
        """Calculates log flows of each state in the trajectories.

        Args:
            env: The environment object.
            trajectories: The batch of trajectories.
            log_pf_trajectories: Tensor of shape (max_length, n_trajectories) containing
                the logprobs of the forward actions of the trajectories.

        Returns:
            A tensor of shape (max_length, n_trajectories) containing the log flows of
            each state in the trajectories.
        """
        states = trajectories.states
        log_state_flows = torch.full_like(log_pf_trajectories, fill_value=-float("inf"))
        mask = ~states.is_sink_state
        valid_states = states[mask]

        if trajectories.conditioning is not None:
            # Compute the conditioning matrix broadcast to match valid_states.
            # The conditioning tensor has shape (max_length, n_trajectories, 1)
            # We need to index it to match the valid states
            conditioning = trajectories.conditioning[mask]

            with has_conditioning_exception_handler("logF", self.logF):
                log_F = self.logF(valid_states, conditioning)
        else:
            with no_conditioning_exception_handler("logF", self.logF):
                log_F = self.logF(valid_states).squeeze(-1)

        if self.forward_looking:
            log_rewards = env.log_reward(states).unsqueeze(-1)
            log_F = log_F + log_rewards

        log_state_flows[mask[:-1]] = log_F.squeeze()
        return log_state_flows

    def calculate_masks(
        self,
        log_state_flows: LogStateFlowsTensor,
        trajectories: Trajectories,
    ) -> Tuple[MaskTensor, MaskTensor]:
        """Calculates masks indicating sink and terminal states.

        Args:
            log_state_flows: Tensor of shape (max_length, n_trajectories) containing
                the log flows of the states.
            trajectories: The batch of trajectories.

        Returns:
            A tuple of two mask tensors (sink_states_mask, is_terminal_mask), each of
            shape (max_length, n_trajectories).
        """
        sink_states_mask = log_state_flows == -float("inf")
        is_terminal_mask = trajectories.actions.is_exit

        return sink_states_mask, is_terminal_mask

    def get_scores(
        self,
        env: Env,
        trajectories: Trajectories,
        recalculate_all_logprobs: bool = True,
    ) -> Tuple[List[torch.Tensor], List[torch.Tensor]]:
        r"""Computes sub-trajectory balance scores for all submitted trajectories.

        Args:
            env: The environment where the trajectories are sampled from.
            trajectories: The batch of trajectories to evaluate.
            recalculate_all_logprobs: Whether to re-evaluate all logprobs.

        Returns:
            A tuple (scores, flattening_masks):
                - scores: A list of tensors, each representing the scores of all
                    sub-trajectories of length k, for k in [1, ..., max_length], where
                    the score of a sub-trajectory $\tau_{n:n+k} = (s_n, ..., s_{n+k})$ is
                    $\log P_F(\tau_{n:n+k}) + \log F(s_n) - \log P_B(\tau_{n:n+k}) - \log F(s_{n+k})$.
                    The shape of each score from k-length sub-trajectory is
                    (max_length - k + 1, n_trajectories).
                - flattening_masks: A list of tensors indicating what should be masked out
                    from the each element of the first list (scores), given that not all
                    sub-trajectories of length k exist for each trajectory. The entries of
                    those tensors are True if the corresponding sub-trajectory does not
                    exist.
        """
        log_pf_trajectories, log_pb_trajectories = self.get_pfs_and_pbs(
            trajectories,
            recalculate_all_logprobs=recalculate_all_logprobs,
        )

        log_pf_trajectories_cum = self.cumulative_logprobs(
            trajectories, log_pf_trajectories
        )
        log_pb_trajectories_cum = self.cumulative_logprobs(
            trajectories, log_pb_trajectories
        )

        log_state_flows = self.calculate_log_state_flows(
            env, trajectories, log_pf_trajectories
        )
        sink_states_mask, is_terminal_mask = self.calculate_masks(
            log_state_flows, trajectories
        )

        flattening_masks = []
        scores = []
        for i in range(1, 1 + trajectories.max_length):
            preds = self.calculate_preds(log_pf_trajectories_cum, log_state_flows, i)
            targets = self.calculate_targets(
                trajectories,
                preds,
                log_pb_trajectories_cum,
                log_state_flows,
                is_terminal_mask,
                sink_states_mask,
                i,
            )

            flattening_mask = trajectories.terminating_idx.lt(
                torch.arange(
                    i,
                    trajectories.max_length + 1,
                    device=trajectories.terminating_idx.device,
                ).unsqueeze(-1)
            )

            flat_preds = preds[~flattening_mask]
            if torch.any(torch.isnan(flat_preds)):
                raise ValueError("NaN in preds")

            flat_targets = targets[~flattening_mask]
            if torch.any(torch.isnan(flat_targets)):
                raise ValueError("NaN in targets")

            flattening_masks.append(flattening_mask)
            scores.append(preds - targets)

        return scores, flattening_masks

    def get_equal_within_contributions(
        self, trajectories: Trajectories
    ) -> ContributionsTensor:
        """Calculates contributions for the 'equal_within' weighting method.

        Args:
            trajectories: The batch of trajectories.

        Returns:
            The contributions tensor of shape (max_len * (max_len+1) / 2, n_trajectories).
        """
        terminating_idx = trajectories.terminating_idx
        max_len = trajectories.max_length
        n_rows = int(max_len * (1 + max_len) / 2)

        # the following tensor represents the inverse of how many sub-trajectories there are in each trajectory
        contributions = (
            2.0 / (terminating_idx * (terminating_idx + 1)) / len(trajectories)
        )

        # if we repeat the previous tensor, we get a tensor of shape
        # (max_len * (max_len + 1) / 2, n_trajectories) that we can multiply with
        # all_scores to get a loss where each sub-trajectory is weighted equally
        # within each trajectory.
        contributions = contributions.repeat(n_rows, 1)

        return contributions

    def get_equal_contributions(self, trajectories: Trajectories) -> ContributionsTensor:
        """Calculates contributions for the 'equal' weighting method.

        Args:
            trajectories: The batch of trajectories.

        Returns:
            The contributions tensor of shape (max_len * (max_len+1) / 2, n_trajectories).
        """
        terminating_idx = trajectories.terminating_idx
        max_len = trajectories.max_length
        n_rows = int(max_len * (1 + max_len) / 2)
        n_sub_trajectories = int(
            (terminating_idx * (terminating_idx + 1) / 2).sum().item()
        )
        contributions = torch.ones(n_rows, len(trajectories)) / n_sub_trajectories
        return contributions

    def get_tb_contributions(self, trajectories: Trajectories) -> ContributionsTensor:
        """Calculates contributions for the 'TB' weighting method.

        Args:
            trajectories: The batch of trajectories.

        Returns:
            The contributions tensor of shape (max_len * (max_len+1) / 2, n_trajectories).
        """
        max_len = trajectories.max_length
        n_rows = int(max_len * (1 + max_len) / 2)
        contributions = torch.zeros(n_rows, len(trajectories))

        # Each trajectory contributes one element to the loss, equally weighted
        terminating_idx = trajectories.terminating_idx
        indices = (
            max_len * (terminating_idx - 1)
            - (terminating_idx - 1) * (terminating_idx - 2) / 2
        ).long()
        contributions.scatter_(0, indices.unsqueeze(0), 1)
        contributions = contributions / len(trajectories)

        return contributions

    def get_modified_db_contributions(
        self, trajectories: Trajectories
    ) -> ContributionsTensor:
        """Calculates contributions for the 'ModifiedDB' weighting method.

        Args:
            trajectories: The batch of trajectories.

        Returns:
            The contributions tensor of shape (max_len * (max_len+1) / 2, n_trajectories).
        """
        terminating_idx = trajectories.terminating_idx
        max_len = trajectories.max_length
        n_rows = int(max_len * (1 + max_len) / 2)

        # The following tensor represents the inverse of how many transitions
        # there are in each trajectory.
        contributions = (1.0 / terminating_idx / len(trajectories)).repeat(max_len, 1)
        contributions = torch.cat(
            (
                contributions,
                torch.zeros(
                    (n_rows - max_len, len(trajectories)),
                    device=contributions.device,
                ),
            ),
            0,
        )
        return contributions

    def get_geometric_within_contributions(
        self, trajectories: Trajectories
    ) -> ContributionsTensor:
        """Calculates contributions for the 'geometric_within' weighting method.

        Args:
            trajectories: The batch of trajectories.

        Returns:
            The contributions tensor of shape (max_len * (max_len+1) / 2, n_trajectories).
        """
        L = self.lamda
        max_len = trajectories.max_length
        terminating_idx = trajectories.terminating_idx

        # The following tensor represents the weights given to each possible
        # sub-trajectory length.
        contributions = (
            L ** torch.arange(max_len, device=terminating_idx.device).double()
        ).to(torch.get_default_dtype())
        contributions = contributions.unsqueeze(-1).repeat(1, len(trajectories))
        contributions = contributions.repeat_interleave(
            torch.arange(max_len, 0, -1, device=terminating_idx.device),
            dim=0,
            output_size=int(max_len * (max_len + 1) / 2),
        )

        # Now we need to divide each column by n + (n-1) lambda +...+ 1*lambda^{n-1}
        # where n is the length of the trajectory corresponding to that column
        # We can do it the ugly way, or using the cool identity:
        # https://www.wolframalpha.com/input?i=sum%28%28n-i%29+*+lambda+%5Ei%2C+i%3D0..n%29
        per_trajectory_denom = (
            1.0
            / (1 - L) ** 2
            * (
                L * (L ** terminating_idx.double() - 1)
                + (1 - L) * terminating_idx.double()
            )
        ).to(torch.get_default_dtype())
        contributions = contributions / per_trajectory_denom / len(trajectories)

        return contributions

    def loss(
        self,
        env: Env,
        trajectories: Trajectories,
        recalculate_all_logprobs: bool = True,
        reduction: str = "mean",
    ) -> torch.Tensor:
        """Computes the sub-trajectory balance loss.

        Args:
            env: The environment where the trajectories are sampled from.
            trajectories: The batch of trajectories to compute the loss with.
            recalculate_all_logprobs: Whether to re-evaluate all logprobs.
            reduction: The reduction method to use ('mean', 'sum', or 'none').

        Returns:
            The computed sub-trajectory balance loss as a tensor. The shape depends on
            the reduction method.
        """
        warn_about_recalculating_logprobs(trajectories, recalculate_all_logprobs)
        # Get all scores and masks from the trajectories.
        scores, flattening_masks = self.get_scores(
            env, trajectories, recalculate_all_logprobs=recalculate_all_logprobs
        )
        flattening_mask = torch.cat(flattening_masks)
        all_scores = torch.cat(scores, 0)

        if self.weighting == "DB":
            # Longer trajectories contribute more to the loss.
            # TODO: is this correct with `loss_reduce`?
            final_scores = scores[0][~flattening_masks[0]].pow(2)
            return loss_reduce(final_scores, reduction)

        elif self.weighting == "geometric":
            # The position i of the following 1D tensor represents the number of sub-
            # trajectories of length i in the batch.
            # n_sub_trajectories = torch.maximum(
            #     trajectories.terminating_idx - torch.arange(3).unsqueeze(-1),
            #     torch.tensor(0),
            # ).sum(1)

            # The following tensor's k-th entry represents the mean of all losses of
            # sub-trajectories of length k.
            per_length_losses = torch.stack(
                [
                    scores[~flattening_mask].pow(2).mean()
                    for scores, flattening_mask in zip(scores, flattening_masks)
                ]
            )
            max_len = trajectories.max_length
            L = self.lamda
            ratio = (1 - L) / (1 - L**max_len)
            weights = ratio * (
                L ** torch.arange(max_len, device=per_length_losses.device)
            )
            assert (weights.sum() - 1.0).abs() < 1e-5, f"{weights.sum()}"
            return (per_length_losses * weights).sum()

        # TODO: we need to know what reductions are valid for each weighting method.
        weight_functions = {
            "equal_within": self.get_equal_within_contributions,
            "equal": self.get_equal_contributions,
            "TB": self.get_tb_contributions,
            "ModifiedDB": self.get_modified_db_contributions,
            "geometric_within": self.get_geometric_within_contributions,
        }
        try:
            contributions = weight_functions[self.weighting](trajectories)
        except KeyError:
            raise ValueError(f"Unknown weighting method {self.weighting}")

        flat_contributions = contributions[~flattening_mask]
        assert (
            flat_contributions.sum() - 1.0
        ).abs() < 1e-5, f"{flat_contributions.sum()}"
        final_scores = flat_contributions * all_scores[~flattening_mask].pow(2)

        # TODO: default was sum, should we allow mean?
        if reduction == "mean":
            warnings.warn(
                "Mean reduction is not supported for SubTBGFlowNet with geometric "
                "weighting, using sum instead."
            )
            reduction = "sum"

        return loss_reduce(final_scores, reduction)



import math
from typing import Tuple

import torch

from gfn.actions import Actions
from gfn.containers import Trajectories, Transitions
from gfn.env import Env
from gfn.estimators import ConditionalScalarEstimator, Estimator, ScalarEstimator
from gfn.gflownet.base import PFBasedGFlowNet, loss_reduce
from gfn.states import States
from gfn.utils.handlers import (
    has_conditioning_exception_handler,
    no_conditioning_exception_handler,
    warn_about_recalculating_logprobs,
)
from gfn.utils.prob_calculations import get_transition_pfs_and_pbs


def check_compatibility(
    states: States, actions: Actions, transitions: Transitions
) -> None:
    """Checks compatibility between states and actions in transitions.

    Args:
        states: The states in the transitions.
        actions: The actions in the transitions.
        transitions: The transitions object.

    Raises:
        TypeError: If transitions is not of type Transitions.
        ValueError: If there is a mismatch between states and actions batch shapes.
    """
    if states.batch_shape != tuple(actions.batch_shape):
        if type(transitions) is not Transitions:
            raise TypeError(
                "`transitions` is type={}, not Transitions".format(type(transitions))
            )
        else:
            raise ValueError(" wrong happening with log_pf evaluations")


class DBGFlowNet(PFBasedGFlowNet[Transitions]):
    r"""GFlowNet for the Detailed Balance loss.

    Corresponds to $\mathcal{O}_{PF} = \mathcal{O}_1 \times \mathcal{O}_2 \times
    \mathcal{O}_3$, where $\mathcal{O}_1$ is the set of functions from the internal
    states (no $s_f$) to $\mathbb{R}^+$ (which we parametrize with logs, to avoid the
    non-negativity constraint), and $\mathcal{O}_2$ is the set of forward probability
    functions consistent with the DAG. $\mathcal{O}_3$ is the set of backward
    probability functions consistent with the DAG, or a singleton thereof, if
    `self.pb` is a fixed `DiscretePBEstimator`.

    The detailed balance loss is described in section
    3.2 of [GFlowNet Foundations](https://arxiv.org/abs/2111.09266).

    Attributes:
        pf: The forward policy estimator.
        pb: The backward policy estimator.
        logF: A ScalarEstimator or ConditionalScalarEstimator for estimating the log
            flow of the states.
        forward_looking: Whether to use the forward-looking GFN loss.
        log_reward_clip_min: If finite, clips log rewards to this value.
        safe_log_prob_min: If True, uses -1e10 as the minimum log probability value
            to avoid numerical instability, otherwise uses -1e38.
        constant_pb: Whether to ignore the backward policy estimator, e.g., if the
            gflownet DAG is a tree, and pb is therefore always 1.
    """

    def __init__(
        self,
        pf: Estimator,
        pb: Estimator | None,
        logF: ScalarEstimator | ConditionalScalarEstimator,
        forward_looking: bool = False,
        log_reward_clip_min: float = -float("inf"),
        safe_log_prob_min: bool = True,
        constant_pb: bool = False,
    ) -> None:
        """Initializes a DBGFlowNet instance.

        Args:
            pf: The forward policy estimator.
            pb: The backward policy estimator, or None if the gflownet DAG is a tree, and
                pb is therefore always 1.
            logF: A ScalarEstimator or ConditionalScalarEstimator for estimating the log
                flow of the states.
            forward_looking: Whether to use the forward-looking GFN loss.
            log_reward_clip_min: If finite, clips log rewards to this value.
            safe_log_prob_min: If True, uses -1e10 as the minimum log probability value
                to avoid numerical instability, otherwise uses -1e38.
            constant_pb: Whether to ignore the backward policy estimator, e.g., if the
                gflownet DAG is a tree, and pb is therefore always 1. Must be set
                explicitly by user to ensure that pb is an Estimator except under this
                special case.
        """
        super().__init__(pf, pb, constant_pb=constant_pb)
        assert any(
            isinstance(logF, cls)
            for cls in [ScalarEstimator, ConditionalScalarEstimator]
        ), "logF must be a ScalarEstimator or derived"
        self.logF = logF
        self.forward_looking = forward_looking
        self.log_reward_clip_min = log_reward_clip_min
        if safe_log_prob_min:
            self.log_prob_min = -1e10
        else:
            self.log_prob_min = -1e38

    def logF_named_parameters(self) -> dict[str, torch.Tensor]:
        """Returns a dictionary of named parameters containing 'logF' in their name.

        Returns:
            A dictionary of named parameters containing 'logF' in their name.
        """
        return {k: v for k, v in self.named_parameters() if "logF" in k}

    def logF_parameters(self) -> list[torch.Tensor]:
        """Returns a list of parameters containing 'logF' in their name.

        Returns:
            A list of parameters containing 'logF' in their name.
        """
        return [v for k, v in self.named_parameters() if "logF" in k]

    def get_pfs_and_pbs(
        self, transitions: Transitions, recalculate_all_logprobs: bool = True
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""Evaluates forward and backward logprobs for each transition in the batch.

        More specifically, it evaluates $\log P_F(s' \mid s)$ and $\log P_B(s \mid s')$
        for each transition in the batch.

        If recalculate_all_logprobs=True, we re-evaluate the logprobs of the transitions
        using the current self.pf. Otherwise, the following applies:
            - If transitions have log_probs attribute, use them - this is usually for
                on-policy learning.
            - Else (transitions have none of them), re-evaluate the logprobs using
                the current self.pf - this is usually for off-policy learning with
                replay buffer.

        Args:
            transitions: The Transitions object to evaluate.
            recalculate_all_logprobs: Whether to re-evaluate all logprobs.

        Returns:
            A tuple of tensors of shape (n_transitions,) containing the log_pf and
            log_pb for each transition.
        """
        return get_transition_pfs_and_pbs(
            self.pf, self.pb, transitions, recalculate_all_logprobs
        )

    def get_scores(
        self, env: Env, transitions: Transitions, recalculate_all_logprobs: bool = True
    ) -> torch.Tensor:
        r"""Calculates the scores for a batch of transitions.

        The scores for each transition are defined as:
        $\log \left( \frac{F(s)P_F(s' \mid s)}{F(s') P_B(s \mid s')} \right)$.

        Args:
            env: The environment where the transitions are sampled from.
            transitions: The Transitions object to evaluate.
            recalculate_all_logprobs: Whether to re-evaluate all logprobs.

        Returns:
            A tensor of shape (n_transitions,) representing the scores for each
            transition.
        """
        if transitions.is_backward:
            raise ValueError("Backward transitions are not supported")

        states = transitions.states
        actions = transitions.actions

        if len(states) == 0:
            return torch.tensor(0.0, device=transitions.device)

        # uncomment next line for debugging
        # assert transitions.states.is_sink_state.equal(transitions.actions.is_dummy)
        check_compatibility(states, actions, transitions)

        log_pf_actions, log_pb_actions = self.get_pfs_and_pbs(
            transitions, recalculate_all_logprobs
        )

        # LogF is potentially a conditional computation.
        if transitions.conditioning is not None:
            with has_conditioning_exception_handler("logF", self.logF):
                log_F_s = self.logF(states, transitions.conditioning).squeeze(-1)
        else:
            with no_conditioning_exception_handler("logF", self.logF):
                log_F_s = self.logF(states).squeeze(-1)

        if self.forward_looking:
            log_rewards = env.log_reward(states)
            if math.isfinite(self.log_reward_clip_min):
                log_rewards = log_rewards.clamp_min(self.log_reward_clip_min)
            log_F_s = log_F_s + log_rewards

        preds = log_pf_actions + log_F_s

        # uncomment next line for debugging
        # assert transitions.next_states.is_sink_state.equal(transitions.is_terminating)

        # automatically removes invalid transitions (i.e. s_f -> s_f)
        valid_next_states = transitions.next_states[~transitions.is_terminating]
        valid_transitions_is_terminating = transitions.is_terminating[
            ~transitions.states.is_sink_state
        ]

        if len(valid_next_states) == 0:
            return torch.tensor(0.0, device=transitions.device)

        # LogF is potentially a conditional computation.
        if transitions.conditioning is not None:
            with has_conditioning_exception_handler("logF", self.logF):
                valid_log_F_s_next = self.logF(
                    valid_next_states,
                    transitions.conditioning[~transitions.is_terminating],
                ).squeeze(-1)
        else:
            with no_conditioning_exception_handler("logF", self.logF):
                valid_log_F_s_next = self.logF(valid_next_states).squeeze(-1)

        log_F_s_next = torch.zeros_like(log_pb_actions)
        log_F_s_next[~valid_transitions_is_terminating] = valid_log_F_s_next
        assert transitions.log_rewards is not None
        valid_transitions_log_rewards = transitions.log_rewards[
            ~transitions.states.is_sink_state
        ]
        log_F_s_next[valid_transitions_is_terminating] = valid_transitions_log_rewards[
            valid_transitions_is_terminating
        ]
        targets = log_pb_actions + log_F_s_next

        scores = preds - targets

        assert scores.shape == (transitions.n_transitions,)
        return scores

    def loss(
        self,
        env: Env,
        transitions: Transitions,
        recalculate_all_logprobs: bool = True,
        reduction: str = "mean",
    ) -> torch.Tensor:
        """Computes the detailed balance loss.

        The detailed balance loss is described in section
        3.2 of [GFlowNet Foundations](https://arxiv.org/abs/2111.09266).

        Args:
            env: The environment where the transitions are sampled from.
            transitions: The Transitions object to compute the loss with.
            recalculate_all_logprobs: Whether to re-evaluate all logprobs.
            reduction: The reduction method to use ('mean', 'sum', or 'none').

        Returns:
            The computed detailed balance loss as a tensor. The shape depends on the
            reduction method.
        """
        warn_about_recalculating_logprobs(transitions, recalculate_all_logprobs)
        scores = self.get_scores(env, transitions, recalculate_all_logprobs)
        scores = scores**2
        loss = loss_reduce(scores, reduction)

        if torch.isnan(loss).any():
            raise ValueError("loss is nan")

        return loss

    def to_training_samples(self, trajectories: Trajectories) -> Transitions:
        """Converts trajectories to transitions for detailed balance loss.

        Args:
            trajectories: The Trajectories object to convert.

        Returns:
            A Transitions object containing all transitions from the trajectories.
        """
        return trajectories.to_transitions()


class ModifiedDBGFlowNet(PFBasedGFlowNet[Transitions]):
    r"""The Modified Detailed Balance GFlowNet.

    Only applicable to environments where all states are terminating. See section 3.2
    of [Bayesian Structure Learning with Generative Flow Networks](https://arxiv.org/abs/2202.13903)
    for more details.

    Attributes:
        pf: The forward policy estimator.
        pb: The backward policy estimator, or None if the gflownet DAG is a tree, and
            pb is therefore always 1.
        constant_pb: Whether to ignore the backward policy estimator, e.g., if the
            gflownet DAG is a tree, and pb is therefore always 1. Must be set explicitly
            by user to ensure that pb is an Estimator except under this special case.
    """

    def __init__(
        self, pf: Estimator, pb: Estimator | None, constant_pb: bool = False
    ) -> None:
        """Initializes a ModifiedDBGFlowNet instance."""
        super().__init__(pf, pb, constant_pb=constant_pb)

    def get_scores(
        self, transitions: Transitions, recalculate_all_logprobs: bool = True
    ) -> torch.Tensor:
        """Calculates DAG-GFN-style modified detailed balance scores.

        Note that this method is only applicable to environments where all states are
        terminating, i.e., the sink state is reachable from all states.

        If recalculate_all_logprobs=True, we re-evaluate the logprobs of the transitions
        using the current self.pf. Otherwise, the following applies:
            - If transitions have log_probs attribute, use them - this is usually for
                on-policy learning.
            - Else, re-evaluate the log_probs using the current self.pf - this is usually
                for off-policy learning with replay buffer.

        Args:
            transitions: The Transitions object to evaluate.
            recalculate_all_logprobs: Whether to re-evaluate all logprobs.

        Returns:
            A tensor of shape (n_transitions,) containing the scores for each transition.
        """
        if transitions.is_backward:
            raise ValueError("Backward transitions are not supported")

        if len(transitions) == 0:
            return torch.tensor(0.0, device=transitions.device)

        mask = ~transitions.next_states.is_sink_state
        states = transitions.states[mask]
        valid_next_states = transitions.next_states[mask]
        actions = transitions.actions[mask]
        all_log_rewards = transitions.all_log_rewards[mask]

        check_compatibility(states, actions, transitions)

        if transitions.conditioning is not None:
            with has_conditioning_exception_handler("pf", self.pf):
                module_output = self.pf(states, transitions.conditioning[mask])
        else:
            with no_conditioning_exception_handler("pf", self.pf):
                module_output = self.pf(states)

        if len(states) == 0:
            return torch.tensor(0.0, device=transitions.device)

        pf_dist = self.pf.to_probability_distribution(states, module_output)

        if transitions.has_log_probs and not recalculate_all_logprobs:
            valid_log_pf_actions = transitions[mask].log_probs
            assert valid_log_pf_actions is not None
        else:
            # Evaluate the log PF of the actions sampled off policy.
            valid_log_pf_actions = pf_dist.log_prob(actions.tensor)
        valid_log_pf_s_exit = pf_dist.log_prob(
            torch.full_like(actions.tensor, actions.__class__.exit_action[0].item())
        )

        # The following two lines are slightly inefficient, given that most
        # next_states are also states, for which we already did a forward pass.
        if transitions.conditioning is not None:
            with has_conditioning_exception_handler("pf", self.pf):
                module_output = self.pf(
                    valid_next_states, transitions.conditioning[mask]
                )
        else:
            with no_conditioning_exception_handler("pf", self.pf):
                module_output = self.pf(valid_next_states)

        valid_log_pf_s_prime_exit = self.pf.to_probability_distribution(
            valid_next_states, module_output
        ).log_prob(
            torch.full_like(actions.tensor, actions.__class__.exit_action[0].item())
        )

        non_exit_actions = actions[~actions.is_exit]

        if self.pb is not None:
            if transitions.conditioning is not None:
                with has_conditioning_exception_handler("pb", self.pb):
                    module_output = self.pb(
                        valid_next_states, transitions.conditioning[mask]
                    )
            else:
                with no_conditioning_exception_handler("pb", self.pb):
                    module_output = self.pb(valid_next_states)

            valid_log_pb_actions = self.pb.to_probability_distribution(
                valid_next_states, module_output
            ).log_prob(non_exit_actions.tensor)
        else:
            # If pb is None, we assume that the gflownet DAG is a tree, and therefore
            # the backward policy probability is always 1 (log probs are 0).
            valid_log_pb_actions = torch.zeros_like(valid_log_pf_s_exit)

        preds = all_log_rewards[:, 0] + valid_log_pf_actions + valid_log_pf_s_prime_exit
        targets = all_log_rewards[:, 1] + valid_log_pb_actions + valid_log_pf_s_exit

        scores = preds - targets
        if torch.any(torch.isinf(scores)):
            raise ValueError("scores contains inf")

        return scores

    def loss(
        self,
        env: Env,
        transitions: Transitions,
        recalculate_all_logprobs: bool = True,
        reduction: str = "mean",
    ) -> torch.Tensor:
        """Computes the modified detailed balance loss.

        Args:
            env: The environment where the transitions are sampled from (unused).
            transitions: The Transitions object to compute the loss with.
            recalculate_all_logprobs: Whether to re-evaluate all logprobs.
            reduction: The reduction method to use ('mean', 'sum', or 'none').

        Returns:
            The computed modified detailed balance loss as a tensor. The shape depends
            on the reduction method.
        """
        del env
        warn_about_recalculating_logprobs(transitions, recalculate_all_logprobs)
        scores = self.get_scores(
            transitions, recalculate_all_logprobs=recalculate_all_logprobs
        )
        scores = scores**2
        return loss_reduce(scores, reduction)

    def to_training_samples(self, trajectories: Trajectories) -> Transitions:
        """Converts trajectories to transitions for modified detailed balance loss.

        Args:
            trajectories: The Trajectories object to convert.

        Returns:
            A Transitions object containing all transitions from the trajectories.
        """
        return trajectories.to_transitions()



"""
The code is adapted from:
https://github.com/GFNOrg/GFN_vs_HVI/blob/master/dags/dag_gflownet/utils/exhaustive.py
"""

from collections import defaultdict, deque
from dataclasses import dataclass
from itertools import permutations
from typing import cast

import networkx as nx
import numpy as np
import pandas as pd
import torch
from pgmpy.estimators import ExhaustiveSearch
from pgmpy.utils.mathext import powerset
from scipy.special import logsumexp
from torch_geometric.data import Data as GeometricData
from tqdm import tqdm

from gfn.actions import GraphActions, GraphActionType
from gfn.estimators import Estimator
from gfn.gym.bayesian_structure import BayesianStructure
from gfn.gym.helpers.bayesian_structure.scores import BaseScore

# https://oeis.org/A003024
NUM_DAGS = [1, 1, 3, 25, 543, 29281, 3781503]


class GraphCollection:
    def __init__(self):
        self.edges, self.lengths = [], []
        self.mapping = defaultdict(int)
        self.mapping.default_factory = lambda: len(self.mapping)

    def append(self, graph: nx.DiGraph) -> None:
        self.edges.extend(
            [self.mapping[edge] for edge in graph.edges()]  # pyright: ignore
        )
        self.lengths.append(graph.number_of_edges())

    def freeze(self) -> "GraphCollection":
        self.edges = np.asarray(self.edges, dtype=np.int32)
        self.lengths = np.asarray(self.lengths, dtype=np.int32)
        self.mapping = [
            edge for (edge, _) in sorted(self.mapping.items(), key=lambda x: x[1])
        ]
        return self

    def is_frozen(self) -> bool:
        return isinstance(self.mapping, list)

    def to_dict(self, prefix: str | None = None) -> dict:
        prefix = f"{prefix}_" if (prefix is not None) else ""
        return {
            f"{prefix}edges": self.edges,
            f"{prefix}lengths": self.lengths,
            f"{prefix}mapping": self.mapping,
        }


@dataclass
class FullPosterior:
    log_probas: np.ndarray
    graphs: GraphCollection
    closures: GraphCollection
    markov: GraphCollection

    def to_dict(self) -> dict:
        # Ensure that "graphs" has been frozen
        if not self.graphs.is_frozen():
            raise ValueError('Graphs must be frozen. Call "graphs.freeze()".')

        offset, output = 0, dict()
        for length, log_prob in zip(self.graphs.lengths, self.log_probas):
            edges_indices = self.graphs.edges[offset : offset + length]
            edges = [self.graphs.mapping[idx] for idx in edges_indices]
            output[frozenset(edges)] = log_prob
            offset += length

        return output


def jensen_shannon_divergence(
    full_posterior: FullPosterior, posterior: FullPosterior
) -> float:
    # Convert to dictionaries to align distributions
    full_posterior_dict = full_posterior.to_dict()
    posterior_dict = posterior.to_dict()

    # Get an (arbitrary ordering of the graphs)
    graphs = list(full_posterior_dict.keys())
    graphs = sorted(graphs, key=len)

    # Get the two distributions aligned
    full_posterior_list, posterior_list = [], []
    for graph in graphs:
        full_posterior_list.append(full_posterior_dict[graph])
        posterior_list.append(posterior_dict[graph])
    full_posterior_arr = np.array(full_posterior_list, dtype=np.float_)
    posterior_arr = np.array(posterior_list, dtype=np.float_)

    # Compute the mean distribution
    mean = np.log(0.5) + np.logaddexp(full_posterior_arr, posterior_arr)

    # Compute the JSD
    KL_full_posterior = np.exp(full_posterior_arr) * (full_posterior_arr - mean)
    KL_posterior = np.exp(posterior_arr) * (posterior_arr - mean)
    return 0.5 * np.sum(KL_full_posterior + KL_posterior)


def get_full_posterior(
    scorer: BaseScore,
    data: pd.DataFrame,
    env: BayesianStructure,
    nodelist: list[str],
    verbose: bool = True,
) -> FullPosterior:
    estimator = ExhaustiveSearch(data, scoring_method=scorer, use_cache=False)

    log_probas = []
    graphs = GraphCollection()
    closures = GraphCollection()
    markov = GraphCollection()

    nx_graphs = list(estimator.all_dags())
    data_array = np.empty(len(nx_graphs), dtype=object)
    for i, graph in enumerate(nx_graphs):
        data_array[i] = nx_to_geometric_data(graph, env, nodelist)
    scores = scorer.state_evaluator(env.States(data_array))
    log_probas = scores.cpu().numpy()
    # Normalize the log-joint distribution to get the posterior
    log_probas -= cast(np.ndarray, logsumexp(log_probas))

    with tqdm(nx_graphs, total=NUM_DAGS[data.shape[1]], disable=(not verbose)) as pbar:
        for graph in pbar:  # Enumerate all possible DAGs
            graphs.append(graph)
            closures.append(nx.transitive_closure_dag(graph))
            markov.append(get_markov_blanket_graph(graph))

    return FullPosterior(
        log_probas=log_probas,
        graphs=graphs.freeze(),
        closures=closures.freeze(),
        markov=markov.freeze(),
    )


def get_gfn_exact_posterior(
    gfn_state_graph: nx.DiGraph, verbose: bool = True
) -> FullPosterior:
    # Get the source graph
    in_degrees = gfn_state_graph.in_degree(gfn_state_graph)
    source_graphs = [
        gfn_state_graph.nodes[node]["graph"]
        for node, in_degree in in_degrees
        if in_degree == 0
    ]
    assert len(source_graphs) == 1
    assert len(source_graphs[0].edges) == 0
    num_variables = len(source_graphs[0])

    log_probas = []
    graphs = GraphCollection()
    closures = GraphCollection()
    markov = GraphCollection()

    for node in tqdm(
        nx.topological_sort(gfn_state_graph),
        total=NUM_DAGS[num_variables],
        disable=(not verbose),
    ):
        graph = gfn_state_graph.nodes[node]["graph"]
        log_probas.append(gfn_state_graph.nodes[node]["terminal_log_flow"])

        graphs.append(graph)
        closures.append(nx.transitive_closure_dag(graph))
        markov.append(get_markov_blanket_graph(graph))

    # The log-posterior is already normalized
    log_probas = np.asarray(log_probas, dtype=np.float_)

    return FullPosterior(
        log_probas,
        graphs=graphs.freeze(),
        closures=closures.freeze(),
        markov=markov.freeze(),
    )


def posterior_exact(
    env: BayesianStructure,
    estimator: Estimator,
    nodelist: list[str],
    batch_size: int = 256,
) -> nx.DiGraph:
    gfn_cache = get_gflownet_cache(env, estimator, nodelist, batch_size)
    gfn_state_graph, source_state_graph = construct_state_dag_with_bfs(
        gfn_cache, nodelist
    )
    gfn_state_graph = push_source_flow_to_terminal_states(
        gfn_state_graph, source_state_graph
    )
    return gfn_state_graph


def get_gflownet_cache(
    env: BayesianStructure,
    estimator: Estimator,
    nodelist: list[str],
    batch_size: int = 256,
) -> dict[frozenset, np.ndarray]:
    """Cache the results of the GFlowNet for all the states.

    This function caches the log-probabilities for all the actions and for
    all the states of the GFlowNet.

    Parameters
    ----------
    env : BayesianStructure
        The Bayesian structure learning environment.
    estimator : Estimator
        The GFlowNet policy estimator.
    nodelist : list[str]
        List of node names to ensure consistent node encoding in adjacency matrices.
    batch_size : int, default=256
        Batch size for processing states through the GFlowNet.

    Returns
    -------
    cache : dict of (frozenset, np.ndarray)
        The cache of log-probabilities returned by the GFlowNet. The keys of
        the cache are the graphs (encoded as a frozenset of their edges), and
        the corresponding value is an array of size `(num_variables ** 2 + 1,)`
        containing the log-probabilities of all the actions in that state
        (including the "stop" action, at the last index).
    """
    cache = dict()

    node_names = {i: name for i, name in enumerate(nodelist)}

    graphs = all_dags(env, len(nodelist), nodelist=nodelist)
    batch_idx = 0
    while batch_idx < len(graphs):
        _graphs = graphs[batch_idx : batch_idx + batch_size]
        _bsz = len(_graphs)
        data_array = np.empty(_bsz, dtype=object)
        for i, graph in enumerate(_graphs):
            data_array[i] = graph

        state = env.States(data_array)
        with torch.no_grad():
            estimator_output = estimator(state)
            dist = estimator.to_probability_distribution(state, estimator_output)

        exit_log_prob = dist.dists[GraphActions.ACTION_TYPE_KEY].log_prob(
            torch.tensor(
                GraphActionType.EXIT, dtype=torch.long, device=env.device
            ).repeat(_bsz)
        )
        exit_log_prob = exit_log_prob.unsqueeze(1)

        add_edge_log_prob = dist.dists[GraphActions.ACTION_TYPE_KEY].log_prob(
            torch.tensor(
                GraphActionType.ADD_EDGE, dtype=torch.long, device=env.device
            ).repeat(_bsz)
        )
        edge_action_log_probs = []
        for edge_index_action in range(len(nodelist) ** 2):
            action_log_prob = dist.dists[GraphActions.EDGE_INDEX_KEY].log_prob(
                torch.tensor(
                    edge_index_action, dtype=torch.long, device=env.device
                ).repeat(_bsz)
            )
            edge_action_log_probs.append(
                (add_edge_log_prob + action_log_prob).unsqueeze(1)
            )
        edge_action_log_probs = torch.cat(edge_action_log_probs, dim=1)

        log_probs = torch.cat([edge_action_log_probs, exit_log_prob], dim=1)

        # TODO: check if no possible edge can be added,
        for graph, log_prob in zip(_graphs, log_probs.cpu().numpy()):
            edges = frozenset(
                (node_names[source.item()], node_names[target.item()])  # pyright: ignore
                for source, target in graph.edge_index.T  # pyright: ignore
            )
            cache[edges] = log_prob

        batch_idx += _bsz

    return cache


def construct_state_dag_with_bfs(
    gflownet_cache: dict[frozenset, np.ndarray],
    nodelist: list[str],
    source_graph: nx.DiGraph | None = None,
) -> tuple[nx.DiGraph, nx.DiGraph]:
    """Constructs the state-action space of the GFlowNet.

    This function performs Breadth-First Search on the GFlowNet state-action space
    starting from the source state, in order to construct a networkx.DiGraph object
    where each node is a GFlowNet state and each edge is labeled with the action
    and the log probability of taking that action. Each node is also labeled with
    the stop_action_log_flow which contains the probability of terminating at that state.

    Parameters
    ----------
    gflownet_cache : dict[frozenset, np.ndarray]
        The cache of log-probabilities returned by the GFlowNet.
    nodelist : list[str]
        The list of nodes.
    source_graph : nx.DiGraph instance
        The graph representing the source state.

    Returns
    -------
    gfn_state_graph : nx.DiGraph instance
        The GFlowNet state-action space.
    source_graph : nx.DiGraph instance
        The graph representing the source state.
    """
    gfn_state_graph = nx.DiGraph()
    is_state_queued = defaultdict(bool)
    states_to_visit = deque()

    if source_graph is None:
        source_graph = nx.DiGraph()
        source_graph.add_nodes_from(nodelist)
    source_graph_key = frozenset(source_graph.edges)

    gfn_state_graph.add_node(source_graph_key, graph=source_graph)
    states_to_visit.append(source_graph)
    is_state_queued[source_graph_key] = True
    while len(states_to_visit) > 0:
        current_graph = states_to_visit.popleft()
        current_graph_key = frozenset(current_graph.edges)
        children = get_children(current_graph, gflownet_cache, nodelist)
        for child_graph, action, log_prob in children:
            if action is None:  # stop action
                # Encode the stop action as a node attribute
                gfn_state_graph.nodes[current_graph_key][
                    "stop_action_log_flow"
                ] = log_prob
            else:
                child_graph_key = frozenset(child_graph.edges)
                if child_graph_key not in gfn_state_graph:
                    gfn_state_graph.add_node(child_graph_key, graph=child_graph)
                gfn_state_graph.add_edge(
                    current_graph_key,
                    child_graph_key,
                    action=action,
                    log_prob_action=log_prob,
                )
                already_visited = is_state_queued[child_graph_key]
                if not already_visited:
                    states_to_visit.append(child_graph)
                    is_state_queued[child_graph_key] = True

    return gfn_state_graph, source_graph


def push_source_flow_to_terminal_states(
    gfn_state_graph: nx.DiGraph,
    source_state_graph: nx.DiGraph,
) -> nx.DiGraph:
    """Compute a hashable key for a graph.

    This function traverses the GFlowNet state-action space graph (DAG) in a
    topologically sorted order and "pushes" the log_flow from each node to
    its children according to the log_prob_action specified on the edges.
    The topological sort ensures that all the flow has "arrived" at a node
    before "moving" its flow to its children.

    Parameters
    ----------
    gfn_state_graph : nx.DiGraph instance
        The GFlowNet state-action space where each node represents one GFlowNet
        state and each edge represents one GFlowNet action.
    source_state_graph: nx.DiGraph instance
        The graph representing the source state.

    Returns
    -------
    gfn_state_graph : nx.DiGraph instance
        The GFlowNet state-action space but now each node has an attribute
        named log_flow, which is -np.inf for non-terminal states and
        the marginal log probability for the terminal states.
    """
    # Initialize log_flow to be -np.inf (flow = 0) for all nodes
    nx.set_node_attributes(gfn_state_graph, -np.inf, "log_flow")  # pyright: ignore

    # Except initialize log_flow to be 0 (flow = 1) for source node
    source_node_key = frozenset(source_state_graph.edges)
    nx.set_node_attributes(gfn_state_graph, {source_node_key: 0}, "log_flow")

    # Push flow through sorted graph
    for state in nx.topological_sort(gfn_state_graph):
        current_node = gfn_state_graph.nodes[state]
        log_flow_incoming = current_node[
            "log_flow"
        ]  # log_flow is log probability of reaching this node starting from source node

        # Compute terminal_log_flow
        stop_action_log_flow = current_node[
            "stop_action_log_flow"
        ]  # probability of taking stop action from this node

        # terminal prob = incoming probability * p(stop action at this node)
        current_node["terminal_log_flow"] = log_flow_incoming + stop_action_log_flow

        # Push flow along edges to children
        edges = gfn_state_graph.edges(state, data=True)
        for _, child, edge_attr in edges:
            log_prob_action = edge_attr["log_prob_action"]
            existing_log_flow_child = gfn_state_graph.nodes[child]["log_flow"]
            updated_log_flow_child = np.logaddexp(
                existing_log_flow_child, log_flow_incoming + log_prob_action
            )
            nx.set_node_attributes(
                gfn_state_graph, {child: updated_log_flow_child}, "log_flow"
            )

    return gfn_state_graph


############################
##### Helper functions #####
############################


def get_markov_blanket(graph: nx.DiGraph, node: str) -> set[str]:
    parents = set(graph.predecessors(node))
    children = set(graph.successors(node))

    mb_nodes = parents | children
    for child in children:
        mb_nodes |= set(graph.predecessors(child))
    mb_nodes.discard(node)

    return mb_nodes


def get_markov_blanket_graph(graph: nx.DiGraph) -> nx.DiGraph:
    """Build an undirected graph where two nodes are connected if
    one node is in the Markov blanket of another.
    """

    def _s(node1: str, node2: str) -> tuple[str, str]:
        return (node2, node1) if (node1 > node2) else (node1, node2)

    # Make it a directed graph to control the order of nodes in each
    # edges, to avoid mapping the same edge to 2 entries in mapping.
    mb_graph = nx.DiGraph()
    mb_graph.add_nodes_from(graph.nodes)

    edges = set()
    for node in graph.nodes:
        edges |= set(_s(node, mb_node) for mb_node in get_markov_blanket(graph, node))
    mb_graph.add_edges_from(edges)

    return mb_graph


def all_dags(
    env: BayesianStructure,
    num_variables: int,
    nodelist: list[str] | None = None,
) -> list[GeometricData]:
    # Adapted from: https://github.com/pgmpy/pgmpy/blob/dev/pgmpy/estimators/ExhaustiveSearch.py
    if nodelist is None:
        nodelist = [str(i) for i in range(num_variables)]

    edges = list(permutations(nodelist, 2))  # n*(n-1) possible directed edges
    all_graphs = powerset(edges)  # 2^(n*(n-1)) graphs

    graphs = []
    for graph_edges in all_graphs:
        graph = nx.DiGraph(graph_edges)  # TODO: use torch_geometric
        graph.add_nodes_from(nodelist)
        if nx.is_directed_acyclic_graph(graph):
            graphs.append(nx_to_geometric_data(graph, env, nodelist))
    return graphs


def nx_to_geometric_data(
    graph: nx.DiGraph, env: BayesianStructure, nodelist: list[str]
) -> GeometricData:
    node_indices = {node: i for i, node in enumerate(nodelist)}
    if graph.edges():
        edge_index = torch.tensor(
            [
                [node_indices[source], node_indices[target]]
                for source, target in graph.edges()
            ],
            dtype=torch.long,
            device=env.device,
        ).T
    else:
        edge_index = torch.zeros(2, 0, dtype=torch.long, device=env.device)

    return GeometricData(
        x=env.s0.x,
        edge_attr=env.s0.edge_attr,
        edge_index=edge_index,
        device=env.device,
    )


def get_valid_actions(graph: nx.DiGraph) -> set[tuple[str, str]]:
    """Gets the list of valid actions.

    The valid actions correspond to directed edges that can be added to the
    current graph, such that adding any of those edges would still yield a
    DAG. In other words, those are edges that (1) are not already present in
    the graph, and (2) would not introduce a directed cycle.

    Parameters
    ----------
    graph : nx.DiGraph instance
        The current graph.

    Returns
    -------
    edges : set of tuples
        A set of directed edges, encoded as a tuple of nodes from `graph`,
        corresponding to the valid actions in the state `graph`.
    """
    if not nx.is_directed_acyclic_graph(graph):
        raise ValueError("The input graph is not a valid DAG.")

    all_edges = set(permutations(graph.nodes, 2))
    edges_already_present = set(graph.edges())

    # Build the transitive closure of the transpose graph
    closure = nx.transitive_closure_dag(graph.reverse())
    edges_cycle = set(closure.edges())

    return all_edges - (edges_already_present | edges_cycle)


def get_children(
    graph: nx.DiGraph,
    gfn_cache: dict[frozenset, np.ndarray],
    nodelist: list[str],
) -> set[tuple[nx.DiGraph, tuple[str, str] | None, np.ndarray]]:
    """Gets all the children of a graph.

    This function returns a set of the next states, from a particular state
    `graph`, with its corresponding log-probability. Note that the set of
    children includes the stop action, encoded as a `None` action, for which
    the child graph is the same as the current graph.

    Parameters
    ----------
    graph : nx.DiGraph instance
        The current graph.

    gfn_cache : dict
        The cache of log-probabilities returned by the GFlowNet. See
        `dag_gflownet.utils.gflownet.get_gflownet_cache` for details.

    nodelist : list
        The list of nodes; this list is required to ensure consistent
        encoding of nodes in the rows and columns of the adjacency matrix.

    Returns
    -------
    children : set of tuples
        The set of all the next state from the current graph, with their
        corresponding log-probability. Each child is represented as
        `(next_graph, action, log_prob)`, where `next_graph` is a nx.DiGraph
        instance, `action` is the edge added (as a tuple of nodes), and
        `log_prob` is the log-probability of this action. Not that the "stop"
        action is encoded as the action `None`.
    """
    node2idx = dict((node, idx) for (idx, node) in enumerate(nodelist))
    valid_actions = get_valid_actions(graph)
    num_variables = len(nodelist)

    log_pi = gfn_cache[frozenset(graph.edges())]
    children: set[tuple[nx.DiGraph, tuple[str, str] | None, np.ndarray]] = {
        (graph, None, log_pi[-1])
    }  # The stop action
    for source, target in valid_actions:
        action = node2idx[source] * num_variables + node2idx[target]
        next_graph = cast(nx.DiGraph, graph.copy())
        next_graph.add_edge(source, target)
        _log_prob = cast(np.ndarray, log_pi[action])
        children.add((next_graph, (source, target), _log_prob))

    return children



from math import log
from typing import Literal, Tuple

import torch

from gfn.actions import Actions
from gfn.env import Env
from gfn.states import States


class Box(Env):
    """Box environment, corresponding to the one in Section 4.1 of https://arxiv.org/abs/2301.12594

    Attributes:
        delta: The step size.
        R0: The base reward.
        R1: The reward for being outside the first box.
        R2: The reward for being inside the second box.
        epsilon: A small value to avoid numerical issues.
        device (Literal["cpu", "cuda"] | torch.device): The device to use.
    """

    def __init__(
        self,
        delta: float = 0.1,
        R0: float = 0.1,
        R1: float = 0.5,
        R2: float = 2.0,
        epsilon: float = 1e-4,
        device: Literal["cpu", "cuda"] | torch.device = "cpu",
        check_action_validity: bool = True,
    ):
        """Initializes the Box environment.

        Args:
            delta: The step size.
            R0: The base reward.
            R1: The reward for being outside the first box.
            R2: The reward for being inside the second box.
            epsilon: A small value to avoid numerical issues.
            device: The device to use.
            check_action_validity: Whether to check the action validity.
        """
        assert 0 < delta <= 1, "delta must be in (0, 1]"
        self.delta = delta
        self.epsilon = epsilon

        s0 = torch.tensor([0.0, 0.0], device=device)
        exit_action = torch.tensor([-float("inf"), -float("inf")], device=device)
        dummy_action = torch.tensor([float("inf"), float("inf")], device=device)

        self.R0 = R0
        self.R1 = R1
        self.R2 = R2

        super().__init__(
            s0=s0,
            state_shape=(2,),  # ()
            action_shape=(2,),
            dummy_action=dummy_action,
            exit_action=exit_action,
            check_action_validity=check_action_validity,
        )

    def make_random_states(
        self, batch_shape: Tuple[int, ...], device: torch.device | None = None
    ) -> States:
        """Generates random states tensor of shape (*batch_shape, 2).

        Args:
            batch_shape: The shape of the batch.
            device: The device to use.

        Returns:
            A States object with random states.
        """
        device = self.device if device is None else device
        return self.States(torch.rand(batch_shape + (2,), device=device))

    def step(self, states: States, actions: Actions) -> States:
        """Step function for the Box environment.

        Args:
            states: States object representing the current states.
            actions: Actions object representing the actions to be taken.

        Returns:
            The next states as a States object.
        """
        return self.States(states.tensor + actions.tensor)

    def backward_step(self, states: States, actions: Actions) -> States:
        """Backward step function for the Box environment.

        Args:
            states: States object representing the current states.
            actions: Actions object representing the actions to be taken.

        Returns:
            The previous states as a States object.
        """
        return self.States(states.tensor - actions.tensor)

    @staticmethod
    def norm(x: torch.Tensor) -> torch.Tensor:
        """Computes the L2 norm of the input tensor along the last dimension.

        Args:
            x: Input tensor of shape `(*batch_shape, 2)`.

        Returns:
            Normalized tensor of shape `batch_shape`.
        """
        return torch.norm(x, dim=-1)

    def is_action_valid(
        self, states: States, actions: Actions, backward: bool = False
    ) -> bool:
        """Checks if the actions are valid.

        Args:
            states: The current states.
            actions: The actions to be taken.
            backward: Whether the actions are backward actions.

        Returns:
            True if the actions are valid, False otherwise.
        """
        non_exit_actions = actions[~actions.is_exit]
        non_terminal_states = states[~actions.is_exit]

        s0_states_idx = non_terminal_states.is_initial_state
        if torch.any(s0_states_idx) and backward:
            return False

        if not backward:
            actions_at_s0 = non_exit_actions[s0_states_idx].tensor

            if torch.any(self.norm(actions_at_s0) > self.delta):
                return False

        non_s0_states = non_terminal_states[~s0_states_idx].tensor
        non_s0_actions = non_exit_actions[~s0_states_idx].tensor

        if (
            not backward
            and torch.any(torch.abs(self.norm(non_s0_actions) - self.delta) > 1e-5)
        ) or torch.any(non_s0_actions < 0):
            return False

        if not backward and torch.any(non_s0_states + non_s0_actions > 1):
            return False

        if backward and torch.any(non_s0_states - non_s0_actions < 0):
            return False

        if backward:
            states_within_delta_radius_idx = self.norm(non_s0_states) < self.delta
            corresponding_actions = non_s0_actions[states_within_delta_radius_idx]
            corresponding_states = non_s0_states[states_within_delta_radius_idx]
            if torch.any(corresponding_actions != corresponding_states):
                return False

        return True

    def reward(self, final_states: States) -> torch.Tensor:
        """Reward is distance from the goal point.

        Args:
            final_states: States object representing the final states.

        Returns:
            The reward tensor of shape `batch_shape`.
        """
        R0, R1, R2 = (self.R0, self.R1, self.R2)
        ax = abs(final_states.tensor - 0.5)
        reward = R0 + (0.25 < ax).prod(-1) * R1 + ((0.3 < ax) * (ax < 0.4)).prod(-1) * R2

        assert reward.shape == final_states.batch_shape
        return reward

    @property
    def log_partition(self) -> float:
        """Returns the log partition of the reward function."""
        return log(self.R0 + (2 * 0.25) ** 2 * self.R1 + (2 * 0.1) ** 2 * self.R2)



import warnings

warnings.warn(
    "'modules.py' is deprecated and will be removed in a future release. Please import "
    "from 'estimators.py' instead.",
    DeprecationWarning,
    stacklevel=2,
)

from gfn.estimators import (  # noqa: F401, E402
    ConditionalDiscretePolicyEstimator,
    ConditionalScalarEstimator,
    DiscreteGraphPolicyEstimator,
    DiscretePolicyEstimator,
    Estimator,
    ScalarEstimator,
)



from __future__ import annotations

from typing import Sequence

import torch

from gfn.actions import Actions
from gfn.containers.base import Container
from gfn.containers.states_container import StatesContainer
from gfn.containers.transitions import Transitions
from gfn.env import Env
from gfn.states import DiscreteStates, GraphStates, States
from gfn.utils.common import ensure_same_device, is_int_dtype


# TODO: remove env from this class?
class Trajectories(Container):
    """Container for complete trajectories (starting in $s_0$ and ending in $s_f$).

    Trajectories are represented as a States object with bi-dimensional batch shape.
    Actions are represented as an Actions object with bi-dimensional batch shape.
    The first dimension represents the time step, the second dimension represents
    the trajectory index. Because different trajectories may have different lengths,
    shorter trajectories are padded with the tensor representation of the terminal
    state ($s_f$ or $s_0$ depending on the direction of the trajectory), and
    actions is appended with dummy actions. The `terminating_idx` tensor represents
    the time step at which each trajectory ends.

    Attributes:
        env: The environment where the states and actions are defined.
        states: States with batch_shape (max_length+1, n_trajectories).
        conditioning: (Optional) Tensor of shape (n_trajectories,) containing the
            conditioning for the trajectories.
        actions: Actions with batch_shape (max_length, n_trajectories).
        terminating_idx: Tensor of shape (n_trajectories,) indicating the time step
            at which each trajectory ends.
        is_backward: Whether the trajectories are backward or forward. When not
            is_backward, the `states` are ordered from initial to terminal states.
            When is_backward, the `states` are ordered from terminal to initial states.
        _log_rewards: (Optional) Tensor of shape (n_trajectories,) containing the
            log rewards of the trajectories.
        log_probs: (Optional) Tensor of shape (max_length, n_trajectories) indicating
            the log probabilities of the trajectories' actions.
        estimator_outputs: (Optional) Tensor of shape (max_length, n_trajectories, ...)
            containing outputs of a function approximator for each step.
    """

    def __init__(
        self,
        env: Env,
        states: States | None = None,
        conditioning: torch.Tensor | None = None,
        actions: Actions | None = None,
        terminating_idx: torch.Tensor | None = None,
        is_backward: bool = False,
        log_rewards: torch.Tensor | None = None,
        log_probs: torch.Tensor | None = None,
        estimator_outputs: torch.Tensor | None = None,
    ) -> None:
        """Initializes a Trajectories instance.

        Args:
            env: The environment where the states and actions are defined.
            states: States with batch_shape (max_length+1, n_trajectories). If None,
                an empty States object is created.
            conditioning: Optional tensor of shape (n_trajectories,) containing the
                conditioning for the trajectories.
            actions: Actions with batch_shape (max_length, n_trajectories). If None,
                an empty Actions object is created.
            terminating_idx: Tensor of shape (n_trajectories,) indicating the time step
                at which each trajectory ends.
            is_backward: Whether the trajectories are backward or forward.
            log_rewards: Optional tensor of shape (n_trajectories,) containing the
                log rewards of the trajectories. If None, computed on the fly when needed.
            log_probs: Optional tensor of shape (max_length, n_trajectories) indicating
                the log probabilities of the trajectories' actions.
            estimator_outputs: Optional tensor of shape (max_length, n_trajectories, ...)
                containing outputs of a function approximator for each step.
        Note:
            When states and actions are not None, the Trajectories is initialized as
            an empty container that can be populated later with the `extend` method.
        """
        self.env = env
        self.is_backward = is_backward

        # Assert that all tensors are on the same device as the environment.
        device = self.env.device
        if isinstance(device, str):
            device = torch.device(device)

        for obj in [states, actions]:
            if obj is not None:
                ensure_same_device(obj.device, device)

        for tensor in [
            conditioning,
            terminating_idx,
            log_rewards,
            log_probs,
            estimator_outputs,
        ]:
            if tensor is not None:
                ensure_same_device(tensor.device, device)

        self.states = (
            states if states is not None else env.states_from_batch_shape((0, 0))
        )
        assert len(self.states.batch_shape) == 2

        self.conditioning = conditioning
        assert self.conditioning is None or (
            self.conditioning.shape[: len(self.states.batch_shape)]
            == self.states.batch_shape
        )

        self.actions = (
            actions if actions is not None else env.actions_from_batch_shape((0, 0))
        )
        assert (self.actions.batch_shape == self.states.batch_shape == (0, 0)) or (
            self.actions.batch_shape
            == (self.states.batch_shape[0] - 1, self.states.batch_shape[1])
        )

        self.terminating_idx = (
            terminating_idx
            if terminating_idx is not None
            else torch.full(size=(0,), fill_value=-1, device=device)
        )
        assert self.terminating_idx.shape == (self.n_trajectories,) and is_int_dtype(
            self.terminating_idx
        )

        self._log_rewards = log_rewards
        assert self._log_rewards is None or (
            self._log_rewards.shape == (self.n_trajectories,)
            and self._log_rewards.is_floating_point()
        )

        self.log_probs = log_probs
        assert self.log_probs is None or (
            self.log_probs.shape == self.actions.batch_shape
            and self.log_probs.is_floating_point()
        )

        self.estimator_outputs = estimator_outputs
        assert self.estimator_outputs is None or (
            self.estimator_outputs.shape[: len(self.states.batch_shape)]
            == self.actions.batch_shape
            and self.estimator_outputs.is_floating_point()
        )

    def __repr__(self) -> str:
        """Returns a string representation of the Trajectories container.

        Returns:
            A string summary of the trajectories.
        """
        trajectories_representation = ""
        n_traj_to_print = min(10, self.n_trajectories)

        if isinstance(self.states, GraphStates):
            for i in range(n_traj_to_print):
                trajectories_representation += str(self.states[..., i]) + "\n"
        else:
            states = self.states.tensor.transpose(0, 1)
            assert states.ndim == 3
            assert isinstance(self.env.s0, torch.Tensor)
            assert isinstance(self.env.sf, torch.Tensor)
            for traj in states[:n_traj_to_print]:
                one_traj_repr = []
                for step in traj:
                    one_traj_repr.append(str(step.cpu().numpy()))
                    if self.is_backward and step.equal(self.env.s0):
                        break
                    elif not self.is_backward and step.equal(self.env.sf):
                        break
                trajectories_representation += "-> ".join(one_traj_repr) + "\n"
        return (
            "Trajectories("
            + f"n_trajectories={self.n_trajectories}, max_length={self.max_length}\n"
            + f"First {n_traj_to_print} trajectories:\n"
            + f"states=\n{trajectories_representation}"
            + f"terminating_idx={self.terminating_idx[:10].cpu().numpy()})"
        )

    @property
    def device(self) -> torch.device:
        """The device on which the trajectories are stored.

        Returns:
            The device object of the `self.states`.
        """
        return self.states.device

    @property
    def n_trajectories(self) -> int:
        """The number of trajectories in the container.

        Returns:
            The number of trajectories.
        """
        return self.states.batch_shape[1]

    def __len__(self) -> int:
        """Returns the number of trajectories in the container.

        Returns:
            The number of trajectories.
        """
        return self.n_trajectories

    @property
    def max_length(self) -> int:
        """The maximum length of the trajectories in the container.

        Returns:
            The maximum trajectory length.
        """
        if len(self) == 0:
            return 0

        return self.actions.batch_shape[0]

    @property
    def terminating_states(self) -> States:
        """The terminating states of the trajectories.

        Returns:
            The terminating states.
        """
        return self.states[self.terminating_idx - 1, torch.arange(self.n_trajectories)]

    @property
    def log_rewards(self) -> torch.Tensor | None:
        """The log rewards for the trajectories.

        Returns:
            Log rewards tensor of shape (n_trajectories,).

        Note:
            If not provided at initialization, log rewards are computed on demand for
            terminating states.
        """
        if self.is_backward:  # TODO: Why can't backward trajectories have log_rewards?
            return None

        if self._log_rewards is None:
            self._log_rewards = self.env.log_reward(self.terminating_states)

        assert self._log_rewards.shape == (self.n_trajectories,)
        return self._log_rewards

    def __getitem__(
        self, index: int | slice | tuple | Sequence[int] | Sequence[bool] | torch.Tensor
    ) -> Trajectories:
        """Returns a subset of the trajectories along the batch dimension.

        Args:
            index: Indices to select trajectories.

        Returns:
            A new Trajectories object with the selected trajectories and associated data.
        """
        if isinstance(index, int):
            index = [index]
        terminating_idx = self.terminating_idx[index]
        new_max_length = terminating_idx.max().item() if len(terminating_idx) > 0 else 0
        states = self.states[:, index]
        conditioning = (
            self.conditioning[:, index] if self.conditioning is not None else None
        )
        actions = self.actions[:, index]
        states = states[: 1 + new_max_length]
        actions = actions[:new_max_length]
        if self.log_probs is not None:
            log_probs = self.log_probs[:, index]
            log_probs = log_probs[:new_max_length]
        else:
            log_probs = None
        log_rewards = self._log_rewards[index] if self._log_rewards is not None else None
        if self.estimator_outputs is not None:
            # TODO: Is there a safer way to index self.estimator_outputs for
            #       for n-dimensional estimator outputs?
            #
            # First we index along the first dimension of the estimator outputs.
            # This can be thought of as the instance dimension, and is
            # compatible with all supported indexing approaches (dim=1).
            # All dims > 1 are not explicitly indexed unless the dimensionality
            # of `index` matches all dimensions of `estimator_outputs` aside
            # from the first (trajectory) dimension.
            estimator_outputs = self.estimator_outputs[:, index]
            # Next we index along the trajectory length (dim=0)
            estimator_outputs = estimator_outputs[:new_max_length]
        else:
            estimator_outputs = None

        return Trajectories(
            env=self.env,
            states=states,
            conditioning=conditioning,
            actions=actions,
            terminating_idx=terminating_idx,
            is_backward=self.is_backward,
            log_rewards=log_rewards,
            log_probs=log_probs,
            estimator_outputs=estimator_outputs,
        )

    def extend(self, other: Trajectories) -> None:
        """Extends this Trajectories object with another Trajectories object.

        Extends along all attributes in turn (actions, states, terminating_idx, log_probs,
        log_rewards).

        Args:
            Another Trajectories to append.
        """
        if self.conditioning is not None:
            # TODO: Support the case
            raise NotImplementedError(
                "`extend` is not implemented for conditional Trajectories."
            )

        if len(other) == 0:
            return

        if len(self) == 0:
            if other._log_rewards is not None:
                self._log_rewards = torch.full(
                    (0,), fill_value=-float("inf"), device=self.device
                )
            if other.log_probs is not None:
                self.log_probs = torch.full(
                    size=(0, 0), fill_value=0.0, device=self.device
                )
            if other.estimator_outputs is not None:
                self.estimator_outputs = torch.full(
                    size=(0, 0, *other.estimator_outputs.shape[2:]),
                    fill_value=0.0,
                    dtype=other.estimator_outputs.dtype,
                    device=self.device,
                )

        # TODO: The replay buffer is storing `dones` - this wastes a lot of space.
        self.actions.extend(other.actions)
        self.states.extend(other.states)  # n_trajectories comes from this.
        self.terminating_idx = torch.cat(
            (self.terminating_idx, other.terminating_idx), dim=0
        )

        # Concatenate log_rewards of the trajectories.
        if self._log_rewards is not None and other._log_rewards is not None:
            self._log_rewards = torch.cat((self._log_rewards, other._log_rewards), dim=0)
            assert len(self._log_rewards) == self.actions.batch_shape[-1]
        else:
            self._log_rewards = None

        # For log_probs, use `pad_dim0_if_needed` to ensure both tensors have the same
        # number of steps (first dimension), padding with 0.0 if necessary, before
        # concatenation.
        if self.log_probs is not None and other.log_probs is not None:
            self.log_probs, other.log_probs = pad_dim0_if_needed(
                self.log_probs, other.log_probs, 0.0
            )
            self.log_probs = torch.cat((self.log_probs, other.log_probs), dim=1)
            assert self.log_probs.shape == self.actions.batch_shape
        else:
            self.log_probs = None

        # Do the same for estimator_outputs, but padding with -float("inf") instead of 0.0
        if self.estimator_outputs is not None and other.estimator_outputs is not None:
            self.estimator_outputs, other.estimator_outputs = pad_dim0_if_needed(
                self.estimator_outputs, other.estimator_outputs
            )
            self.estimator_outputs = torch.cat(
                (self.estimator_outputs, other.estimator_outputs), dim=1
            )
            assert (
                self.estimator_outputs.shape[: len(self.actions.batch_shape)]
                == self.actions.batch_shape
            )
        else:
            self.estimator_outputs = None

    def to_transitions(self) -> Transitions:
        """Returns a Transitions object from the current Trajectories.

        Returns:
            A Transitions object with the same states, actions, and log_rewards as the
            current Trajectories.
        """
        if self.conditioning is not None:
            # The conditioning tensor has shape (max_length, n_trajectories, 1)
            # The actions have shape (max_length, n_trajectories)
            # We need to index the conditioning tensor to match the actions
            # The actions exclude the last step, so we need to exclude the last step from conditioning
            conditioning = self.conditioning[:-1][~self.actions.is_dummy]
        else:
            conditioning = None

        states = self.states[:-1][~self.actions.is_dummy]
        next_states = self.states[1:][~self.actions.is_dummy]
        actions = self.actions[~self.actions.is_dummy]
        is_terminating = (
            next_states.is_sink_state
            if not self.is_backward
            else next_states.is_initial_state
        )
        if self._log_rewards is None:
            log_rewards = None
        else:
            log_rewards = torch.full(
                actions.batch_shape,
                fill_value=-float("inf"),
                device=actions.device,
            )
            # TODO: Can we vectorize this?
            log_rewards[is_terminating] = torch.cat(
                [
                    self._log_rewards[self.terminating_idx == i]
                    for i in range(self.terminating_idx.max() + 1)
                ],
                dim=0,
            )
        log_probs = (
            self.log_probs[~self.actions.is_dummy]
            if self.log_probs is not None
            else None
        )

        return Transitions(
            env=self.env,
            states=states,
            conditioning=conditioning,
            actions=actions,
            is_terminating=is_terminating,
            next_states=next_states,
            is_backward=self.is_backward,
            log_rewards=log_rewards,
            log_probs=log_probs,
            # FIXME: Add estimator_outputs.
        )

    def to_states_container(self) -> StatesContainer:
        """Returns a StatesContainer object from the current Trajectories.

        Returns:
            A StatesContainer object with the same states, actions, and log_rewards as the
            current Trajectories.
        """
        if not isinstance(self.states, DiscreteStates):
            raise TypeError("to_states_container only works with DiscreteStates")

        is_terminating = torch.zeros(
            self.states.batch_shape, dtype=torch.bool, device=self.states.device
        )
        is_terminating[self.terminating_idx - 1, torch.arange(len(self))] = True

        states = self.states.flatten()
        is_terminating = is_terminating.flatten()

        is_valid = ~states.is_sink_state & (
            ~states.is_initial_state | (states.is_initial_state & is_terminating)
        )
        states = states[is_valid]
        is_terminating = is_terminating[is_valid]

        conditioning = None
        if self.conditioning is not None:
            # The conditioning tensor has shape (max_length, n_trajectories, 1)
            # We need to flatten it to match the flattened states
            # First, we need to repeat it to match the flattened shape
            # The flattened states have shape (max_length * n_trajectories,)
            # So we need to repeat the conditioning tensor accordingly
            conditioning = self.conditioning.flatten(0, 1)[is_valid]

        if self.log_rewards is None:
            log_rewards = None
        else:
            log_rewards = torch.full(
                (len(states),),
                fill_value=-float("inf"),
                device=states.device,
            )
            # Get the original indices (before flattening and filtering).
            orig_batch_indices = torch.arange(
                self.states.batch_shape[0], device=states.device
            ).repeat_interleave(self.states.batch_shape[1])
            orig_traj_indices = torch.arange(
                self.states.batch_shape[1], device=states.device
            ).repeat(self.states.batch_shape[0])

            # Retain only the valid indices.
            valid_batch_indices = orig_batch_indices[is_valid]
            valid_traj_indices = orig_traj_indices[is_valid]

            # Assign rewards to valid terminating states.
            terminating_mask = is_terminating & (
                valid_batch_indices == (self.terminating_idx[valid_traj_indices] - 1)
            )
            log_rewards[terminating_mask] = self.log_rewards[
                valid_traj_indices[terminating_mask]
            ]

        return StatesContainer[DiscreteStates](
            env=self.env,
            states=states,
            conditioning=conditioning,
            is_terminating=is_terminating,
            log_rewards=log_rewards,
            # FIXME: Add log_probs and estimator_outputs.
        )

    def reverse_backward_trajectories(self) -> Trajectories:
        """Returns a reversed version of the backward trajectories."""
        assert self.is_backward, "Trajectories must be backward."
        # env.sf should never be None unless something went wrong during class
        # instantiation.
        if self.env.sf is None:
            raise AttributeError(
                "Something went wrong during the instantiation of environment {}".format(
                    self.env
                )
            )

        # Compute sequence lengths and maximum length
        seq_lengths = self.terminating_idx  # shape (n_trajectories,)
        max_len = int(seq_lengths.max().item())

        # Get actions and states
        actions = self.actions  # shape (max_len, n_trajectories *action_dim)
        states = self.states  # shape (max_len + 1, n_trajectories, *state_dim)

        # Initialize new actions and states
        new_actions = self.env.Actions.make_dummy_actions(
            (max_len + 1, len(self)), device=actions.device
        )
        # shape (max_len + 1, n_trajectories, *action_dim)
        new_states = self.env.States.make_sink_states(
            (max_len + 2, len(self)), device=states.device
        )
        # shape (max_len + 2, n_trajectories, *state_dim)

        # Create helper indices and masks
        idx = (
            torch.arange(max_len, device=seq_lengths.device)
            .unsqueeze(1)
            .expand(-1, len(self))
        )

        rev_idx = seq_lengths.unsqueeze(0) - 1 - idx  # shape (max_len, n_trajectories)
        mask = rev_idx >= 0  # shape (max_len, n_trajectories)

        # -------------------------------------------------------------
        # Replace the previous transpose-based reversal logic with a
        # version that operates directly in (time, trajectory, *) space.
        # -------------------------------------------------------------

        # 1. Reverse actions ---------------------------------------------------
        # Gather linear indices where the mask is valid
        time_idx, traj_idx = torch.nonzero(mask, as_tuple=True)  # 1-D tensors
        src_time_idx = rev_idx[mask]  # Corresponding source time indices
        # Assign reversed actions
        new_actions[time_idx, traj_idx] = actions[src_time_idx, traj_idx]
        # Insert EXIT action right after the last real action of every trajectory
        new_actions[seq_lengths, torch.arange(len(self), device=seq_lengths.device)] = (
            self.env.Actions.make_exit_actions((1,), device=actions.device)
        )

        # 2. Reverse states ----------------------------------------------------
        # The last state of the backward trajectories must be s0.
        assert torch.all(states[-1].is_initial_state), "Last state must be s0"

        # First state of the forward trajectories is s0 for every trajectory
        new_states[0] = self.env.States.make_initial_states(
            (len(self),), device=states.device
        )  # Broadcast over the trajectory dimension

        # We do not want to copy the last state (s0) from the backward trajectory.
        states_excl_last = states[:-1]  # shape (max_len, n_trajectories, *state_dim)
        new_states_data = new_states[1:-1]  # shape (max_len, n_trajectories, *state_dim)
        new_states_data[time_idx, traj_idx] = states_excl_last[src_time_idx, traj_idx]

        # ---------------------------------------------------------------------
        # new_actions / new_states already have the correct shapes
        #   new_actions: (max_len + 1, n_trajectories, *action_dim)
        #   new_states:  (max_len + 2, n_trajectories, *state_dim)
        # ---------------------------------------------------------------------

        reversed_trajectories = Trajectories(
            env=self.env,
            states=new_states,
            conditioning=self.conditioning,
            actions=new_actions,
            terminating_idx=self.terminating_idx + 1,
            is_backward=False,
            log_rewards=self.log_rewards,
            log_probs=None,  # We can't simply pass the trajectories.log_probs
            # Since `log_probs` is assumed to be the forward log probabilities.
            # FIXME: To resolve this, we can save log_pfs and log_pbs in the
            # trajectories object.
            estimator_outputs=None,  # Same as `log_probs`.
        )

        return reversed_trajectories


def pad_dim0_if_needed(
    a: torch.Tensor, b: torch.Tensor, value: float = -float("inf")
) -> tuple[torch.Tensor, torch.Tensor]:
    """Pads tensor a or b to match the first dimension of the other.

    Args:
        a: First tensor.
        b: Second tensor.
        value: Value to use for padding.

    Returns:
        Tuple of tensors with the same first dimension.
    """
    if a.shape[0] == b.shape[0]:
        return a, b

    max_dim0 = max(a.shape[0], b.shape[0])

    tensor_to_pad = a if a.shape[0] < b.shape[0] else b

    pad_dim = max_dim0 - min(a.shape[0], b.shape[0])
    pad_dim_full = (pad_dim,) + tuple(tensor_to_pad.shape[1:])
    output_padding = torch.full(
        pad_dim_full,
        fill_value=value,
        dtype=tensor_to_pad.dtype,
        device=tensor_to_pad.device,
    )

    tensor_to_pad = torch.cat((tensor_to_pad, output_padding), dim=0)

    return (
        tensor_to_pad if a.shape[0] < b.shape[0] else a,
        tensor_to_pad if a.shape[0] > b.shape[0] else b,
    )



from __future__ import annotations  # This allows to use the class name in type hints

import enum
from abc import ABC
from math import prod
from typing import ClassVar, List, Sequence

import torch
from tensordict import TensorDict


class Actions(ABC):
    """Base class for actions, representing edges in the DAG of a GFlowNet.

    Each environment needs to define a subclass of `Actions` to represent its specific
    action space.

    Two useful subclasses of `Actions` are provided:
    - `DiscreteActions` for discrete environments, which represents actions as a tensor
      of shape (*batch_shape, *action_shape).
    - `GraphActions` for graph-based environments, which represents actions as a tensor
      of shape (*batch_shape, 4) containing the action type, node class, edge class,
      and edge index components.

    Attributes:
        tensor: Tensor of shape (*batch_shape, *action_shape) representing a batch of
            actions.
        action_shape: Class variable, a tuple defining the shape of a single action.
        dummy_action: Class variable, a tensor of shape (*action_shape,) representing
            the dummy action for padding shorter trajectories.
        exit_action: Class variable, a tensor of shape (*action_shape,) representing
            the action to transition to the sink state.
    """

    # The following class variable represents the shape of a single action.
    action_shape: ClassVar[tuple[int, ...]]  # All actions need to have the same shape.
    # The following class variable is padded to shorter trajectories.
    dummy_action: ClassVar[torch.Tensor]  # Dummy action for the environment.
    # The following class variable corresponds to $s \rightarrow s_f$ transitions.
    exit_action: ClassVar[torch.Tensor]  # Action to exit the environment.

    def __init__(self, tensor: torch.Tensor):
        """Initializes an Actions object with a batch of actions.

        Args:
            tensor: Tensor of shape (*batch_shape, *action_shape) representing a batch of
                actions.
        """
        assert (
            tensor.shape[-len(self.action_shape) :] == self.action_shape
        ), f"Batched actions tensor has shape {tensor.shape}, but the expected action shape is {self.action_shape}."

        self.tensor = tensor

    @property
    def device(self) -> torch.device:
        """The device on which the actions are stored.

        Returns:
            The device of the underlying tensor.
        """
        return self.tensor.device

    @property
    def batch_shape(self) -> tuple[int, ...]:
        """The batch shape of the actions.

        Returns:
            The batch shape as a tuple.
        """
        return tuple(self.tensor.shape)[: -len(self.action_shape)]

    @classmethod
    def make_dummy_actions(
        cls, batch_shape: tuple[int, ...], device: torch.device | None = None
    ) -> Actions:
        """Creates an Actions object filled with dummy actions.

        Args:
            batch_shape: Shape of the batch dimensions.
            device: The device to create the actions on.

        Returns:
            An Actions object with the specified batch shape filled with dummy actions.
        """
        action_ndim = len(cls.action_shape)
        tensor = cls.dummy_action.repeat(*batch_shape, *((1,) * action_ndim))
        if device is not None:
            tensor = tensor.to(device)
        return cls(tensor)

    @classmethod
    def make_exit_actions(
        cls, batch_shape: tuple[int, ...], device: torch.device | None = None
    ) -> Actions:
        """Creates an Actions object filled with exit actions.

        Args:
            batch_shape: Shape of the batch dimensions.
            device: The device to create the actions on.

        Returns:
            An Actions object with the specified batch shape filled with exit actions.
        """
        action_ndim = len(cls.action_shape)
        tensor = cls.exit_action.repeat(*batch_shape, *((1,) * action_ndim))
        if device is not None:
            tensor = tensor.to(device)
        return cls(tensor)

    def __len__(self) -> int:
        """Returns the number of actions in the batch.

        Returns:
            The number of actions.
        """
        return prod(self.batch_shape)

    def __repr__(self):
        """Returns a string representation of the Actions object.

        Returns:
            A string summary of the Actions object.
        """
        parts = [
            f"{self.__class__.__name__}(",
            f"batch={self.batch_shape},",
            f"action={self.action_shape},",
            f"device={self.device})",
        ]
        return " ".join(parts)

    def __getitem__(
        self, index: int | slice | tuple | Sequence[int] | Sequence[bool] | torch.Tensor
    ) -> Actions:
        """Returns a subset of the actions along the batch dimension.

        Args:
            index: Indices to select actions.

        Returns:
            A new Actions object with the selected actions.
        """
        actions = self.tensor[index]
        return self.__class__(actions)

    def __setitem__(
        self,
        index: int | slice | tuple | Sequence[int] | Sequence[bool] | torch.Tensor,
        actions: Actions,
    ) -> None:
        """Sets particular actions of the batch to a new Actions object.

        Args:
            index: Indices to set.
            actions: Actions object containing the new actions.
        """
        self.tensor[index] = actions.tensor

    @classmethod
    def stack(cls, actions_list: List[Actions]) -> Actions:
        """Stacks a list of Actions objects along a new dimension (0).

        The individual actions need to have the same batch shape. An example application
        is when the individual actions represent per-step actions of a batch of
        trajectories (in which case, the common batch_shape would be (n_trajectories,),
        and the resulting Actions object would have batch_shape
        (n_steps, n_trajectories)).

        Args:
            actions_list: List of Actions objects to stack.

        Returns:
            A new Actions object with the stacked actions.
        """
        actions_tensor = torch.stack([actions.tensor for actions in actions_list], dim=0)
        return cls(actions_tensor)

    def extend(self, other: Actions) -> None:
        """Concatenates another Actions object along the final batch dimension.

        Both Actions objects must have the same number of batch dimensions, which
        should be 1 or 2.

        Args:
            other: Actions object to be concatenated to the current Actions object.
        """
        if len(self.batch_shape) == len(other.batch_shape) == 1:
            self.tensor = torch.cat((self.tensor, other.tensor), dim=0)
        elif len(self.batch_shape) == len(other.batch_shape) == 2:
            self.extend_with_dummy_actions(
                required_first_dim=max(self.batch_shape[0], other.batch_shape[0])
            )
            other.extend_with_dummy_actions(
                required_first_dim=max(self.batch_shape[0], other.batch_shape[0])
            )
            self.tensor = torch.cat((self.tensor, other.tensor), dim=1)
        else:
            raise NotImplementedError(
                "extend is only implemented for bi-dimensional actions."
            )

    def extend_with_dummy_actions(self, required_first_dim: int) -> None:
        """Extends an Actions instance along the first dimension with dummy actions.

        The Actions instance batch_shape must be 2-dimensional. This is used to pad
        actions in a batch of trajectories to a common length.

        Args:
            required_first_dim: The target size of the first dimension post expansion.
        """
        if len(self.batch_shape) == 2:
            if self.batch_shape[0] >= required_first_dim:
                return
            n = required_first_dim - self.batch_shape[0]
            dummy_actions = self.__class__.make_dummy_actions(
                (n, self.batch_shape[1]), device=self.device
            )
            self.tensor = torch.cat((self.tensor, dummy_actions.tensor), dim=0)
        else:
            raise NotImplementedError(
                "extend_with_dummy_actions is only implemented for bi-dimensional actions."
            )

    def _compare(self, other: torch.Tensor) -> torch.Tensor:
        """Compares the actions to a tensor of actions.

        Args:
            other: Tensor of actions to compare, with shape (*batch_shape, *action_shape).

        Returns:
            A boolean tensor of shape (*batch_shape,) indicating whether the actions are
            equal.
        """
        n_batch_dims = len(self.batch_shape)
        if n_batch_dims == 1:
            assert (other.shape == self.action_shape) or (
                other.shape == self.batch_shape + self.action_shape
            ), f"Expected shape {self.action_shape} or {self.batch_shape + self.action_shape}, got {other.shape}."
        else:
            assert (
                other.shape == self.batch_shape + self.action_shape
            ), f"Expected shape {self.batch_shape + self.action_shape}, got {other.shape}."

        out = self.tensor == other
        if len(self.action_shape) > 1:
            out = out.flatten(start_dim=n_batch_dims)
        out = out.all(dim=-1)

        assert out.shape == self.batch_shape
        return out

    @property
    def is_dummy(self) -> torch.Tensor:
        """Returns a boolean tensor indicating whether the actions are dummy actions.

        Returns:
            A boolean tensor of shape (*batch_shape,) that is True for dummy actions.
        """
        if len(self.batch_shape) == 1:
            dummy_actions_tensor = self.__class__.dummy_action
        else:
            dummy_actions_tensor = self.__class__.dummy_action.repeat(
                *self.batch_shape, *((1,) * len(self.__class__.action_shape))
            )
        return self._compare(dummy_actions_tensor)

    @property
    def is_exit(self) -> torch.Tensor:
        """Returns a boolean tensor indicating whether the actions are exit actions.

        Returns:
            A boolean tensor of shape (*batch_shape,) that is True for exit actions.
        """
        if len(self.batch_shape) == 1:
            exit_actions_tensor = self.__class__.exit_action
        else:
            exit_actions_tensor = self.__class__.exit_action.repeat(
                *self.batch_shape, *((1,) * len(self.__class__.action_shape))
            )
        return self._compare(exit_actions_tensor)


class GraphActionType(enum.IntEnum):
    ADD_NODE = 0
    ADD_EDGE = 1
    EXIT = 2
    DUMMY = 3


class GraphActions(Actions):
    """Actions for graph-based environments.

    Each action is one of these types:
    - ADD_NODE: Add a node with given features
    - ADD_EDGE: Add an edge between two nodes with given features
    - EXIT: Terminate the trajectory

    Attributes:
        tensor: Tensor of shape (*batch_shape, 4) containing the action type, node class,
            edge class, and edge index components.
        ACTION_TYPE_KEY: Class variable, key for the action type component.
        NODE_CLASS_KEY: Class variable, key for the node class component.
        EDGE_CLASS_KEY: Class variable, key for the edge class component.
        EDGE_INDEX_KEY: Class variable, key for the edge index component.
        ACTION_INDICES: Class variable, mapping from keys to tensor indices.
    """

    ACTION_TYPE_KEY: ClassVar[str] = "action_type"
    NODE_CLASS_KEY: ClassVar[str] = "node_class"
    EDGE_CLASS_KEY: ClassVar[str] = "edge_class"
    EDGE_INDEX_KEY: ClassVar[str] = "edge_index"

    ACTION_INDICES: ClassVar[dict[str, int]] = {
        ACTION_TYPE_KEY: 0,
        NODE_CLASS_KEY: 1,
        EDGE_CLASS_KEY: 2,
        EDGE_INDEX_KEY: 3,
    }

    def __init__(self, tensor: torch.Tensor):
        """Initializes a GraphActions object.

        Args:
            tensor: A tensor of shape (*batch_shape, 4) containing the action type,
                node class, edge class, and edge index components.
        """
        if tensor.shape[-1] != 4:
            raise ValueError(
                f"Expected tensor of shape (*batch_shape, 4), got {tensor.shape}.\n"
                "The last dimension should contain the action type, node class, edge class, and edge index."
            )
        self.tensor = tensor

    @property
    def batch_shape(self) -> tuple[int, ...]:
        """The batch shape of the graph actions.

        Returns:
            The batch shape as a tuple.
        """
        assert self.tensor.shape[-1] == 4
        return self.tensor.shape[:-1]

    @classmethod
    def from_tensor_dict(cls, tensor_dict: TensorDict) -> GraphActions:
        """Creates a GraphActions object from a tensor dict.

        Args:
            tensor_dict: A TensorDict containing the action components with keys
                ACTION_TYPE_KEY, NODE_CLASS_KEY, EDGE_CLASS_KEY, and EDGE_INDEX_KEY.

        Returns:
            A GraphActions object constructed from the tensor dict.
        """
        batch_shape = tensor_dict[cls.ACTION_TYPE_KEY].shape
        action_type = tensor_dict[cls.ACTION_TYPE_KEY].reshape(*batch_shape, 1)
        node_class = tensor_dict[cls.NODE_CLASS_KEY].reshape(*batch_shape, 1)
        edge_class = tensor_dict[cls.EDGE_CLASS_KEY].reshape(*batch_shape, 1)
        edge_index = tensor_dict[cls.EDGE_INDEX_KEY].reshape(*batch_shape, 1)

        return cls(torch.cat([action_type, node_class, edge_class, edge_index], dim=-1))

    def __repr__(self):
        """Returns a string representation of the GraphActions object.

        Returns:
            A string summary of the GraphActions object.
        """
        return f"""GraphAction object with {self.batch_shape} actions."""

    @property
    def is_exit(self) -> torch.Tensor:
        """Returns a boolean tensor indicating whether the actions are exit actions.

        Returns:
            A boolean tensor of shape (*batch_shape,) that is True for exit actions.
        """
        return self.action_type == GraphActionType.EXIT

    @property
    def is_dummy(self) -> torch.Tensor:
        """Returns a boolean tensor indicating whether the actions are dummy actions.

        Returns:
            A boolean tensor of shape (*batch_shape,) that is True for dummy actions.
        """
        return self.action_type == GraphActionType.DUMMY

    @property
    def action_type(self) -> torch.Tensor:
        """Returns the action type tensor.

        Returns:
            A tensor of shape (*batch_shape,) containing the action types.
        """
        return self.tensor[..., self.ACTION_INDICES[self.ACTION_TYPE_KEY]]

    @property
    def node_class(self) -> torch.Tensor:
        """Returns the node class tensor.

        Returns:
            A tensor of shape (*batch_shape,) containing the node classes.
        """
        return self.tensor[..., self.ACTION_INDICES[self.NODE_CLASS_KEY]]

    @property
    def edge_class(self) -> torch.Tensor:
        """Returns the edge class tensor.

        Returns:
            A tensor of shape (*batch_shape,) containing the edge classes.
        """
        return self.tensor[..., self.ACTION_INDICES[self.EDGE_CLASS_KEY]]

    @property
    def edge_index(self) -> torch.Tensor:
        """Returns the edge index tensor.

        Returns:
            A tensor of shape (*batch_shape,) containing the edge indices.
        """
        return self.tensor[..., self.ACTION_INDICES[self.EDGE_INDEX_KEY]]

    @classmethod
    def make_dummy_actions(
        cls, batch_shape: tuple[int], device: torch.device
    ) -> GraphActions:
        """Creates a GraphActions object filled with dummy actions.

        Args:
            batch_shape: Shape of the batch dimensions.
            device: The device to create the actions on.

        Returns:
            A GraphActions object with the specified batch shape filled with dummy
            actions.
        """
        tensor = torch.zeros(batch_shape + (4,), dtype=torch.long, device=device)
        tensor[..., cls.ACTION_INDICES[cls.ACTION_TYPE_KEY]] = GraphActionType.DUMMY
        return cls(tensor)

    @classmethod
    def make_exit_actions(
        cls, batch_shape: tuple[int], device: torch.device
    ) -> GraphActions:
        """Creates a GraphActions object filled with exit actions.

        Args:
            batch_shape: Shape of the batch dimensions.
            device: The device to create the actions on.

        Returns:
            A GraphActions object with the specified batch shape filled with exit actions.
        """
        tensor = torch.zeros(batch_shape + (4,), dtype=torch.long, device=device)
        tensor[..., cls.ACTION_INDICES[cls.ACTION_TYPE_KEY]] = GraphActionType.EXIT
        return cls(tensor)

    @classmethod
    def edge_index_action_to_src_dst(
        cls, edge_index_action: torch.Tensor, n_nodes: int
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Converts the edge index action to source and destination node indices."""
        raise NotImplementedError("Not implemented.")



from typing import Tuple

import torch

from gfn.containers import Trajectories, Transitions
from gfn.estimators import Estimator
from gfn.states import States
from gfn.utils.handlers import (
    has_conditioning_exception_handler,
    no_conditioning_exception_handler,
)


def check_cond_forward(
    module: Estimator,
    module_name: str,
    states: States,
    condition: torch.Tensor | None = None,
) -> torch.Tensor:
    """Checks if conditioning is passed and calls the module's forward method accordingly.

    Args:
        module: The GFN module to call.
        module_name: The name of the module (for error messages).
        states: The states to pass to the module.
        condition: Optional conditioning tensor to pass to the module.

    Returns:
        The output of the module's forward method.

    Raises:
        TypeError: If conditioning is passed but the module does not accept it, or vice-versa.
    """
    if condition is not None:
        with has_conditioning_exception_handler(module_name, module):
            return module(states, condition)
    else:
        with no_conditioning_exception_handler(module_name, module):
            return module(states)


# ------------
# Trajectories
# ------------


def get_trajectory_pfs_and_pbs(
    pf: Estimator,
    pb: Estimator | None,
    trajectories: Trajectories,
    fill_value: float = 0.0,
    recalculate_all_logprobs: bool = True,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Calculates the log probabilities of forward and backward trajectories.

    Args:
        pf: The forward policy estimator.
        pb: The backward policy estimator, or None if the gflownet DAG is a tree, and
            pb is therefore always 1.
        trajectories: The trajectories to calculate probabilities for.
        fill_value: The value to fill for invalid states (e.g., sink states).
        recalculate_all_logprobs: Whether to recalculate log probabilities even if they
            already exist in the trajectories object.

    Returns:
        A tuple containing two tensors: log_pf_trajectories and log_pb_trajectories.
    """
    # fill value is the value used for invalid states (sink state usually)

    # uncomment next line for debugging
    # assert trajectories.states.is_sink_state[:-1].equal(trajectories.actions.is_dummy)

    log_pf_trajectories = get_trajectory_pfs(
        pf,
        trajectories,
        fill_value=fill_value,
        recalculate_all_logprobs=recalculate_all_logprobs,
    )
    log_pb_trajectories = get_trajectory_pbs(pb, trajectories, fill_value=fill_value)

    return log_pf_trajectories, log_pb_trajectories


def get_trajectory_pfs(
    pf: Estimator,
    trajectories: Trajectories,
    fill_value: float = 0.0,
    recalculate_all_logprobs: bool = True,
) -> torch.Tensor:
    """Calculates the log probabilities of forward trajectories.

    Args:
        pf: The forward policy estimator.
        trajectories: The trajectories to calculate probabilities for.
        fill_value: The value to fill for invalid states (e.g., sink states).
        recalculate_all_logprobs: Whether to recalculate log probabilities even if they
            already exist in the trajectories object.

    Returns:
        A tensor containing the log probabilities of the forward trajectories.

    Raises:
        ValueError: If backward trajectories are provided.
    """
    if trajectories.is_backward:
        raise ValueError("Backward trajectories are not supported")

    state_mask = ~trajectories.states.is_sink_state
    action_mask = ~trajectories.actions.is_dummy

    valid_states = trajectories.states[state_mask]
    valid_actions = trajectories.actions[action_mask]

    if valid_states.batch_shape != valid_actions.batch_shape:
        raise AssertionError("Something wrong happening with log_pf evaluations")

    if trajectories.has_log_probs and not recalculate_all_logprobs:
        log_pf_trajectories = trajectories.log_probs
        assert log_pf_trajectories is not None
    else:
        log_pf_trajectories = torch.full_like(
            trajectories.actions.tensor[..., 0],
            fill_value=fill_value,
            dtype=torch.get_default_dtype(),  # Floating point dtype.
        )

        if len(valid_states) == 0:
            return log_pf_trajectories

        if trajectories.estimator_outputs is not None and not recalculate_all_logprobs:
            estimator_outputs = trajectories.estimator_outputs[action_mask]
        else:
            masked_cond = None
            if trajectories.conditioning is not None:
                cond_dim = (-1,) * len(trajectories.conditioning.shape)
                traj_len = trajectories.states.tensor.shape[0]
                masked_cond = trajectories.conditioning.unsqueeze(0).expand(
                    (traj_len,) + cond_dim
                )[state_mask]

            estimator_outputs = check_cond_forward(pf, "pf", valid_states, masked_cond)

        # Calculates the log PF of the actions sampled off policy.
        valid_log_pf_actions = pf.to_probability_distribution(
            valid_states, estimator_outputs
        ).log_prob(
            valid_actions.tensor
        )  # Using the actions sampled off-policy.

        log_pf_trajectories[action_mask] = valid_log_pf_actions

    assert log_pf_trajectories.shape == (
        trajectories.max_length,
        trajectories.n_trajectories,
    )

    return log_pf_trajectories


def get_trajectory_pbs(
    pb: Estimator | None,
    trajectories: Trajectories,
    fill_value: float = 0.0,
) -> torch.Tensor:
    """Calculates the log probabilities of backward trajectories.

    Args:
        pb: The backward policy estimator.
        trajectories: The trajectories to calculate probabilities for.
        fill_value: The value to fill for invalid states (e.g., sink states).
        dtype: The dtype of the log probabilities.

    Returns:
        A tensor containing the log probabilities of the backward trajectories.

    Raises:
        ValueError: If backward trajectories are provided.
    """
    if trajectories.is_backward:
        raise ValueError("Backward trajectories are not supported")

    log_pb_trajectories = torch.full_like(
        trajectories.actions.tensor[..., 0],
        fill_value=fill_value,
        dtype=torch.get_default_dtype(),  # Floating point dtype.
    )

    # Note the different mask for valid states and actions compared to the pf case.
    state_mask = (
        ~trajectories.states.is_sink_state & ~trajectories.states.is_initial_state
    )
    # We can't calculate the PB of the first state, even it's not an initial state.
    state_mask[0, :] = False
    action_mask = ~trajectories.actions.is_dummy & ~trajectories.actions.is_exit

    valid_states = trajectories.states[state_mask]
    valid_actions = trajectories.actions[action_mask]

    if valid_states.batch_shape != valid_actions.batch_shape:
        raise AssertionError("Something wrong happening with log_pf evaluations")

    if len(valid_states) == 0:
        return log_pb_trajectories

    # Using all non-initial states, calculate the backward policy, and the logprobs
    # of those actions.
    masked_cond = None
    if trajectories.conditioning is not None:
        # We need to index the conditioning vector to broadcast over the states.
        # The conditioning tensor has shape (max_length, n_trajectories, 1)
        # We need to index it with the state_mask to get the valid states
        masked_cond = trajectories.conditioning[state_mask]

    if pb is not None:
        estimator_outputs = check_cond_forward(pb, "pb", valid_states, masked_cond)
        valid_log_pb_actions = pb.to_probability_distribution(
            valid_states, estimator_outputs
        ).log_prob(valid_actions.tensor)

    else:
        # If pb is None, we assume that the gflownet DAG is a tree, and therefore
        # the backward policy probability is always 1 (log probs are 0).
        valid_log_pb_actions = torch.zeros_like(valid_actions.tensor)

    log_pb_trajectories[action_mask] = valid_log_pb_actions

    assert log_pb_trajectories.shape == (
        trajectories.max_length,
        trajectories.n_trajectories,
    )

    return log_pb_trajectories


# -----------
# Transitions
# -----------


def get_transition_pfs_and_pbs(
    pf: Estimator,
    pb: Estimator | None,
    transitions: Transitions,
    recalculate_all_logprobs: bool = True,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Calculates the log probabilities of forward and backward transitions.

    Args:
        pf: The forward policy estimator.
        pb: The backward policy estimator, or None if the gflownet DAG is a tree, and
            pb is therefore always 1.
        transitions: The transitions to calculate probabilities for.
        recalculate_all_logprobs: Whether to recalculate log probabilities even if they
            already exist in the transitions object.

    Returns:
        A tuple containing two tensors: log_pf_transitions and log_pb_transitions.

    Raises:
        ValueError: If backward transitions are provided.
    """
    if transitions.is_backward:
        raise ValueError("Backward transitions are not supported")

    log_pf_transitions = get_transition_pfs(pf, transitions, recalculate_all_logprobs)
    log_pb_transitions = get_transition_pbs(pb, transitions)

    assert log_pf_transitions.shape == (transitions.n_transitions,)
    assert log_pb_transitions.shape == (transitions.n_transitions,)

    return log_pf_transitions, log_pb_transitions


def get_transition_pfs(
    pf: Estimator, transitions: Transitions, recalculate_all_logprobs: bool = True
) -> torch.Tensor:
    """Calculates the log probabilities of forward transitions.

    Args:
        pf: The forward policy estimator.
        transitions: The transitions to calculate probabilities for.
        recalculate_all_logprobs: Whether to recalculate log probabilities even if they
            already exist in the transitions object.

    Returns:
        A tensor containing the log probabilities of the forward transitions.
    """
    states = transitions.states
    actions = transitions.actions

    if transitions.has_log_probs and not recalculate_all_logprobs:
        log_pf_actions = transitions.log_probs
        assert log_pf_actions is not None
    else:
        # Evaluate the log PF of the actions, with optional conditioning.
        # TODO: Inefficient duplication in case of tempered policy
        # The Transitions container should then have some
        # estimator_outputs attribute as well, to avoid duplication here ?
        # See (#156).
        estimator_outputs = check_cond_forward(
            pf, "pf", states, transitions.conditioning
        )

        log_pf_actions = pf.to_probability_distribution(
            states, estimator_outputs
        ).log_prob(actions.tensor)

    return log_pf_actions


def get_transition_pbs(pb: Estimator | None, transitions: Transitions) -> torch.Tensor:
    """Calculates the log probabilities of backward transitions.

    Args:
        pb: The backward policy Estimator, or None if the gflownet DAG is a tree, and
            pb is therefore always 1.
        transitions: The transitions to calculate probabilities for.
    """
    # automatically removes invalid transitions (i.e. s_f -> s_f)
    valid_next_states = transitions.next_states[~transitions.is_terminating]
    non_exit_actions = transitions.actions[~transitions.actions.is_exit]

    # Evaluate the log PB of the actions, with optional conditioning.
    masked_cond = (
        transitions.conditioning[~transitions.is_terminating]
        if transitions.conditioning is not None
        else None
    )

    # TODO: We support a fill_value for trajectories, but not for transitions.
    # Should we add it here, or remove it for trajectories?
    log_pb_actions = torch.zeros(
        (transitions.n_transitions,), device=transitions.states.device
    )

    # If pb is None, we assume that the gflownet DAG is a tree, and therefore
    # the backward policy probability is always 1 (log probs are 0).
    if pb is not None:
        estimator_outputs = check_cond_forward(pb, "pb", valid_next_states, masked_cond)

        # Evaluate the log PB of the actions.
        valid_log_pb_actions = pb.to_probability_distribution(
            valid_next_states, estimator_outputs
        ).log_prob(non_exit_actions.tensor)

        if len(valid_next_states) != 0:
            log_pb_actions[~transitions.is_terminating] = valid_log_pb_actions

    return log_pb_actions



import math
from abc import ABC, abstractmethod

import torch
from scipy.special import gammaln


class BasePrior(ABC):
    """Base class for the prior over graphs p(G).

    Any subclass of `BasePrior` must return the contribution of log p(G) for a
    given variable with `num_parents` parents. We assume that the prior is modular.

    Parameters
    ----------
    num_variables : int (optional)
        The number of variables in the graph. If not specified, this gets
        populated inside the scorer class.
    """

    def __init__(self, num_variables=None):
        self._num_variables = num_variables
        self._log_prior = None

    def __call__(self, num_parents):
        if self.log_prior is None:
            raise ValueError("log_prior has not been initialized.")
        return self.log_prior[num_parents]

    @property
    @abstractmethod
    def log_prior(self):
        pass

    @property
    def num_variables(self):
        if self._num_variables is None:
            raise RuntimeError("The number of variables is not defined.")
        return self._num_variables

    @num_variables.setter
    def num_variables(self, value):
        self._num_variables = value


class UniformPrior(BasePrior):
    @property
    def log_prior(self):
        self._log_prior = torch.zeros(self.num_variables)
        return self._log_prior


class ErdosRenyiPrior(BasePrior):
    def __init__(self, num_variables=None, num_edges_per_node=1.0):
        super().__init__(num_variables)
        self.num_edges_per_node = num_edges_per_node

    @property
    def log_prior(self):
        num_edges = self.num_variables * self.num_edges_per_node  # Default value
        p = num_edges / ((self.num_variables * (self.num_variables - 1)) // 2)
        all_parents = torch.arange(self.num_variables)
        self._log_prior = all_parents * math.log(p) + (
            self.num_variables - all_parents - 1
        ) * math.log1p(-p)
        return self._log_prior


class EdgePrior(BasePrior):
    def __init__(self, num_variables=None, beta=1.0):
        super().__init__(num_variables)
        self.beta = beta

    @property
    def log_prior(self):
        self._log_prior = torch.arange(self.num_variables) * math.log(self.beta)
        return self._log_prior


class FairPrior(BasePrior):
    @property
    def log_prior(self):
        all_parents = torch.arange(self.num_variables)
        self._log_prior = (
            -gammaln(self.num_variables + 1)
            + gammaln(self.num_variables - all_parents + 1)
            + gammaln(all_parents + 1)
        )
        return self._log_prior



from .base import Container
from .replay_buffer import NormBasedDiversePrioritizedReplayBuffer, ReplayBuffer
from .states_container import StatesContainer
from .trajectories import Trajectories
from .transitions import Transitions

__all__ = [
    "NormBasedDiversePrioritizedReplayBuffer",
    "ReplayBuffer",
    "StatesContainer",
    "Trajectories",
    "Transitions",
    "Container",
]
